####################################################################################################
# neuropythy/datasets/hcp.py
# The HCP_1200 dataset
# by Noah C. Benson

import os, six, shutil, tarfile, logging, warnings, pimms
import numpy as np
import pyrsistent as pyr

from six.moves import urllib

if six.PY3: from functools import reduce

from .core        import (Dataset, add_dataset)
from ..util       import (config, curry, auto_dict, to_credentials, pseudo_path)
from ..vision     import as_retinotopy
from ..           import io      as nyio
from .. import hcp

config.declare_credentials('hcp_credentials',
                           environ_name='HCP_CREDENTIALS',
                           extra_environ=[('HCP_KEY', 'HCP_SECRET'),
                                          'S3FS_CREDENTIALS',
                                          ('S3FS_KEY', 'S3FS_SECRET')],
                           filenames=['~/.hcp-passwd', '~/.passwd-hcp',
                                      '~/.s3fs-passwd', '~/.passwd-s3fs'],
                           aws_profile_name=['HCP', 'hcp', 'S3FS', 's3fs'])
def to_nonempty(s):
    '''
    to_nonempty(s) yields s if s is a nonempty string and otherwise raises an exception.
    '''
    if not pimms.is_str(s) or s == '': raise ValueError('cannot convert object to non-empty string')
    return s
def to_nonempty_path(s):
    '''
    to_nonempty_path(s) yields s if s is a nonempty string and otherwise raises an exception. If s
      is a string, then the variable- and user-expanded form is returned.
    '''
    if not pimms.is_str(s) or s == '': raise ValueError('cannot convert object to non-empty string')
    return os.path.expanduser(os.path.expandvars(s))
config.declare('hcp_auto_release', environ_name='HCP_AUTO_RELEASE', default_value='HCP_1200',
               filter=to_nonempty)
config.declare('hcp_auto_database',environ_name='HCP_AUTO_DATABASE',default_value='hcp-openaccess',
               filter=to_nonempty)
config.declare('hcp_auto_path',    environ_name='HCP_AUTO_PATH',    default_value=Ellipsis,
               filter=to_nonempty_path)
config.declare('hcp_auto_default_alignment',
               environ_name='HCP_AUTO_DEFAULT_ALIGNMENT',
               default_value='MSMAll',
               filter=to_nonempty)
def to_interpolation_method(im):
    '''
    to_interpolation_method(x) yields either 'linear' or 'nearest' or raises an exception, depending
      on x; if x is None or a string resembling 'nearest', then 'nearest' is returned, while a
      string resembling 'linear' results in 'linear' being returned.
    '''
    if im is None: return 'nearest'
    im = im.lower()
    if   im in ['linear','lin','trilinear','trilin']: return 'linear'
    elif im in ['nearest','near','nn','nearest-neighbor','nearest_neighbor']: return 'nearest'
    else: raise ValueError('given interpolation_method (%s) is not recognized' % (im,))
config.declare('hcp_retinotopy_interpolation_method',
               environ_name='HCP_RETINOTOPY_INTERPOLATION_METHOD',
               default_value='nearest', filter=to_interpolation_method)
# See if the config/environment lets auto-downloading start in the "on" state
def to_boolean(arg):
    '''
    to_boolean(arg) yields True or False based on the given arg. This function is intended to
      convert a neuropythy config argument into a boolean value. Accordingly, it accepts True or
      False values, numbers, or the strings "true", "on", "yes", "1", etc. as true and "false",
      "off", etc. as false. If arg has no clear True or False value (such as an arbitrary non-empty
      string) then None is returned.
    '''
    # Basically, we want to do a boolean test, but there are some special cases.
    if   arg is None: return False
    elif arg is Ellipsis: return None
    elif pimms.is_str(arg):
        arg = arg.strip().lower()
        return (True  if arg in ('true',  't', 'yes', '1',  'on', '+') else
                False if arg in ('false', 'f',  'no', '0', 'off',  '') else
                None)
    elif pimms.is_number(arg):
        return None if np.isnan(arg) else (arg != 0)
    elif pimms.is_nparray(arg):
        # Handle numpy arrays specially, because they don't like being bool'ed.
        return np.prod(arg.shape) > 0
    else:
        return bool(arg)
def to_auto_download_state(arg):
    '''
    to_auto_download_state(arg) attempts to coerce the given argument into a valid auto-downloading
      instruction. Essentially, if arg is "on", "yes", "true", "1", True, or 1, then True is
      returned; if arg is "structure" then "structure" is returned; otherwise False is returned. The
      special value Ellipsis yields True if the 'hcp_credentials' have been provided and False
      otherwise.
    '''
    if arg is Ellipsis: return config['hcp_credentials'] is not None
    else:               return to_boolean(arg)
config.declare('hcp_auto_download', environ_name='HCP_AUTO_DOWNLOAD',
               filter=to_auto_download_state, default_value=Ellipsis)

# HHCPMetaDataset ##################################################################################
config.declare('hcp_metadata_path', environ_name='HCP_METADATA_PATH', default_value=None,
               filter=to_nonempty_path)
config.declare('hcp_behavioral_path', environ_name='HCP_BEHAVIORAL_PATH', default_value=None,
               filter=to_nonempty_path)
config.declare('hcp_genetic_path', environ_name='HCP_GENETIC_PATH', default_value=None,
               filter=to_nonempty_path)
@pimms.immutable
class HCPMetaDataset(Dataset):
    '''
    The HCPMetaDataset (in neuropythy.data['hcp_metadata'] is a common repository for the various
    behavioral/genetic/meta-data provided by the Human Connectome Project. Because these data are
    not generally available for auto-download and/or are behind a registration process, you have
    to obtain these data manually then configure neuropythy to know about them. Neuropythy does not
    actually use these itself, but other datasets can use them and simply require you to configure
    the dataset through this interface instead of telling each dataset individually where it should
    find the files. A couple of summaries (gender and agegroup) are also provided here.

    To tell neuropythy where to find the relevant files, you use the configuration interface,
    ideally by making a JSON file named .npythyrc in your home directory that contains the paths for
    the relevant files. These files and configuration names are described below. If any of these
    paths is omitted from the configuration, there are a couple of default places that neuropythy
    will look for files with particular names. These are in the "hcp_subjects_path", in the
    "cache_data_root", and in the "hcp_metadata" and "hcp/metadata" subdirectories of the 
    "cache_data_root" configuration item. Both "hcp_subject_paths" and "cache_data_root" are
    configured via neuropythy.config as well and are documented elsewhere. If multiple HCP subject
    paths are included, then all of them are searched. Finally, the optional configuration item,
    "hcp_metadata_path" can be set to a directory which should be searched for files with the
    default names.

      * "hcp_behavioral_path" should be the path of the behavioral data CSV file provided by the
        Human Connectome Project (https://db.humanconnectome.org/). The default search name for
        this file is behvaioral_data_1200_subjects.csv.
      * "hcp_genetic_path" should be the path of the restricted family-structure data CSV file
        provided by the Human Connectome Project. The default search name for this file is the
        "RESTRICTED_hcpfamilystructure.csv"

    If you have declared the object `data = neuropythy.data['hcp_metadata']` then the following
    are the relevant members of data.
      * data.
    '''
    def __init__(self, name='hcp_metadata', cache_directory=None,
                 metadata_path=None, genetic_path=None, behavioral_path=None,
                 create_mode=0o755, create_directories=False,
                 meta_data=None, cache_required=False):
        '''
        HCPMetaDataset() creates a new HCP meta-dataset object. Constructing a new object will
        reset the object based on the current state of the neuropythy config data; alternately,
        values for metadata_path, genetic_path, and behavioral_path may be passed.
        '''
        Dataset.__init__(self, name,
                         meta_data=meta_data,
                         custom_directory=cache_directory,
                         create_directories=create_directories,
                         create_mode=create_mode,
                         cache_required=cache_required)
        self.metadata_path = metadata_path
        self._genetic_path = genetic_path
        self._behavioral_path = behavioral_path
    # How we search for a file
    @staticmethod
    def _findfile(paths, name):
        import os
        for path in paths:
            if path is None or path is Ellipsis: continue
            path = os.path.expanduser(os.path.expandvars(path))
            fname = os.path.join(path, name)
            if os.path.isfile(fname): return fname
        return None
    @pimms.param
    def metadata_path(p):
        '''
        metadata_path is either None (when no metadata-path has been provided/configured) or the
        path to the meta-data directory of the HCP. This directory is simply a location in which
        neuropythy will look for behavioral and genetic metadata about HCP subjects.
        '''
        import os
        if p is None or p is Ellipsis:
            # We want to see if it's in the config (this is None if not)
            return config['hcp_metadata_path']
        else:
            p = os.path.expanduser(os.path.expandvars(p))
            if not os.path.isdir(p):
                raise ValueError('metadata_path must be a directory')
            return p
    @pimms.param
    def _behavioral_path(p):
        '''
        _behavioral_path is the path to the behavioral/meta-data CSV file provided by the Human
        Connectome Project. This is the argument given to the dataset, not the actual name of
        the file used (which may be found elsewhere if None is given, for example).
        '''
        import os
        if p is None or p is Ellipsis: return None
        if not pimms.is_str(p): raise ValueError('behavioral_path must be None or a string')
        return os.path.expanduser(os.path.expandvars(p))
    @pimms.param
    def _genetic_path(p):
        '''
        _genetic_path is the path to the family-structure data CSV file provided by the Human
        Connectome Project. This is the argument given to the dataset, not the actual name of
        the file used (which may be found elsewhere if None is given, for example).
        '''
        import os
        if p is None or p is Ellipsis: return None
        if not pimms.is_str(p): raise ValueError('genetic_path must be None or a string')
        return os.path.expanduser(os.path.expandvars(p))
    @pimms.value
    def behavioral_path(_behavioral_path, metadata_path):
        '''
        behavioral_path is the path of the file being used for behavioral data. If no such file
        is found this is None.
        '''
        if _behavioral_path is not None:
            if not os.path.isfile(_behavioral_path):
                raise ValueError('provided behavioral_path is not a file')
            return _behavioral_path
        f = config['hcp_behavioral_path']
        if f is not None:
            if os.path.isfile(f): return f
            else: warnings.warn('provided config item hcp_behavioral_path is not a file')
        # we look through a few default places
        hcps = config['hcp_subject_paths']
        hcps = [] if hcps is None else hcps
        paths = [metadata_path] + hcps + [config['data_cache_root']]
        f = HCPMetaDataset._findfile(paths, 'behavioral_data_1200_subjects.csv')
        return f
    @pimms.value
    def genetic_path(_genetic_path, metadata_path):
        '''
        genetic_path is the path of the file being used for the restricted family-structure data.
        If no such file is found this is None.
        '''
        if _genetic_path is not None:
            if not os.path.isfile(_genetic_path):
                raise ValueError('provided genetic_path is not a file')
            return _genetic_path
        f = config['hcp_genetic_path']
        if f is not None:
            if os.path.isfile(f): return f
            else: warnings.warn('provided config item hcp_genetic_path is not a file')
        # we look through a few default places
        hcps = config['hcp_subject_paths']
        hcps = [] if hcps is None else hcps
        paths = [metadata_path] + hcps + [config['data_cache_root']]
        f = HCPMetaDataset._findfile(paths, 'RESTRICTED_hcpfamilystructure.csv')
        return f
    @pimms.value
    def behavioral_table(behavioral_path):
        '''
        behavioral_table is a pandas datafame of the behavioral data provided by the Human
        Connectome Project.
        '''
        if behavioral_path is None: return None
        return nyio.load(behavioral_path)
    @pimms.value
    def genetic_table(genetic_path):
        '''
        genetic_table is a pandas dataframe of the restricted genetic data for all subjects from
        the HCP.
        '''
        if genetic_path is None: return None
        return nyio.load(genetic_path)
    @pimms.value
    def behavioral_maps(behavioral_table):
        '''
        behavioral_maps is a dictionary of meta-data dictionaries, one per subject by subject ID.
        '''
        from pyrsistent import pmap
        from neuropythy.util import dataframe_select
        if behavioral_table is None: return None
        mdat = {}
        sids = np.unique(behavioral_table['Subject'].values)
        for sid in sids:
            tmp = dataframe_select(behavioral_table, Subject=sid)
            mdat[sid] = pmap(dict(tmp.iloc[0]))
        return pmap(mdat)
    @pimms.value
    def gender(behavioral_maps):
        '''
        gender is a dictionary of genders ('F' or 'M') for each subject, based on the behavioral
        data provided by the HCP.
        '''
        from pyrsistent import pmap
        from six import iteritems
        if behavioral_maps is None: return None
        return pmap({sid: v['Gender'] for (sid,v) in iteritems(behavioral_maps)})
    @pimms.value
    def agegroup(behavioral_maps):
        '''
        agegroup is a dictionary of agegroup numbers, as provided by the HCP's behavioral data.
        The agegroup number is the mean of the highest and lowest age in the agegroup except for
        the '36+' agegroup, which is coded as a 40.
        '''
        from pyrsistent import pmap
        from six import iteritems
        if behavioral_maps is None: return None
        agegroup = pyr.pmap(
            {sid: np.mean([int(a1), int(a2)])
             for (sid,v) in iteritems(behavioral_maps)
             for age in [v['Age']]
             for (a1,a2) in [age.split('-') if age != '36+' else ['40','40']]})
        return pmap(agegroup)
    @pimms.value
    def siblings(genetic_table):
        '''
        siblings is a mapping of the siblings in the HCP restricted genetics dataset.
        '''
        families = {}
        for (ii,r) in genetic_table.iterrows():
            (sid,fid,zyg) = [r[k] for k in ('Subject','Family_ID','ZygosityGT')]
            zyg = zyg.strip()
            if fid not in families: families[fid] = {}
            fam = families[fid]
            if zyg not in fam: fam[zyg] = []
            fam[zyg].append(sid)
        siblings = {'':{}, 'MZ':{}, 'DZ':{}}
        for (ii,r) in genetic_table.iterrows():
            (sid,fid,zyg) = [r[k] for k in ('Subject','Family_ID','ZygosityGT')]
            zyg = zyg.strip()
            sib = siblings[zyg]
            rel = families[fid][zyg]
            sib[sid] = [k for k in rel if k != sid]
        # clean the siblings up (twins have only one entry so clear the lists)
        for tw in ['DZ','MZ']: siblings[tw] = {k:v[0] for (k,v) in six.iteritems(siblings[tw])}
        siblings[''] = {k:v for (k,v) in six.iteritems(siblings['']) if len(v) > 0}
        return pimms.persist(siblings)
    @pimms.value
    def retinotopy_siblings(siblings):
        '''
        retinotopy_siblings is a mapping like siblings but restricted to just the subjects with
        retinotopic maps.
        '''
        # make the retinotopy subset of subjects:
        slist = HCPRetinotopyDataset.subject_ids
        retinotopy_siblings = {
            kk: {k:v for (k,v) in six.iteritems(vv)
                 if k in slist
                 if all(u in slist for u in ([v] if pimms.is_int(v) else v))}
            for (kk,vv) in six.iteritems(siblings)}
        return pimms.persist(retinotopy_siblings)
    @staticmethod
    def _siblings_to_pairs(rs):
        subject_list = [u for v in six.itervalues(rs)
                        for uuu in [[six.iterkeys(v)], six.itervalues(v)]
                        for uu in uuu for u in ([uu] if pimms.is_int(uu) else uu)]
        subject_list = np.unique(subject_list)
        # setup twin numbers so that we can export anonymized twin data (i.e.,
        # files containing twin data but not the subject IDs)
        twin_pairs = {tw: pimms.imm_array(list(sorted(dat)))
                      for tw  in ['MZ','DZ']
                      for dat in [set([tuple(sorted([k,v])) for (k,v) in six.iteritems(rs[tw])])]}
        # also get a list of all siblings so we can track who is/isn't related
        siblings = {}
        for s1 in subject_list:
            q = []
            for sibs in six.itervalues(rs):
                if s1 not in sibs: continue
                ss = sibs[s1]
                if pimms.is_int(ss): ss = [ss]
                for s2 in ss: q.append(s2)
            if len(q) > 0: siblings[s1] = q
        # Make up a list of all possible unrelated pairs
        unrelated_pairs = []
        for sid in subject_list:
            # find a random subject to pair them with
            urs = np.setdiff1d(subject_list, [sid] + siblings.get(sid,[]))
            unrelated_pairs.append([urs, np.full(len(urs), sid)])
        unrelated_pairs = np.unique(np.sort(np.hstack(unrelated_pairs), axis=0), axis=1).T
        unrelated_pairs.setflags(write=False)
        # Having made those unrelated pairs, we can add them to the twin pairs
        twin_pairs['UR'] = unrelated_pairs
        # finally, let's figure out the non-twin siblings:
        sibs = [(k,v) for (k,vv) in six.iteritems(rs['']) for v in vv]
        twin_pairs['SB'] = np.unique(np.sort(sibs, axis=1), axis=0)
        twin_pairs['SB'].setflags(write=False)
        return pyr.pmap({'monozygotic_twins': twin_pairs['MZ'],
                         'dizygotic_twins':   twin_pairs['DZ'],
                         'nontwin_siblings':  twin_pairs['SB'],
                         'unrelated_pairs':   twin_pairs['UR']})
    @pimms.value
    def sibling_pairs(siblings):
        '''
        sibling_pairs is a persistent map of twin pairs, sibling pairs, and unrelated
        pairs; each of these categories stores a (n x 2) matrix of the n pairs of subjects
        associated with that category.

        The keys are 'monozygotic_twins', 'dizygotic_twins', 'nontwin_siblings',
        and 'unrelated_pairs'.
        '''
        return HCPMetaDataset._siblings_to_pairs(siblings)
    @pimms.value
    def retinotopy_sibling_pairs(retinotopy_siblings):
        '''
        retinotopy_sibling_pairs is a persistent map of twin pairs, sibling pairs, and
        unrelated pairs; each of these categories stores a (n x 2) matrix of the n pairs of subjects
        associated with that category. This dataset is equivalent to sibling_pairs, except that it
        is restricted to subjects with retinotopic mapping data.

        The keys are 'monozygotic_twins', 'dizygotic_twins', 'nontwin_siblings',
        and 'unrelated_pairs'.
        '''
        return HCPMetaDataset._siblings_to_pairs(retinotopy_siblings)
add_dataset('hcp_metadata', lambda:HCPMetaDataset().persist())

# HCPDataset #######################################################################################
@pimms.immutable
class HCPDataset(HCPMetaDataset):
    '''
    neuropythy.data['hcp'] is a Dataset containing the publicly provided structural data from the
    Human Connectome Project; see https://db.humanconnectome.org/ for more information about the
    dataset itself.

    In order to use this dataset, you will need to configure a few items; primarily, you will need
    to provide some credentials via the neuropythy.config interface (see also:
    https://github.com/noahbenson/neuropythy/wiki/Configuration). You can generate HCP S3
    credentials at the db.humanconnectome.org site; once these are generated, they should either be
    placed in your neuropythy config file or set manually:
      `ny.config['hcp_credentials'] = 'dummykey:dummysecret'`

    You will probably also want to specify a location to cache the HCP subject data; if this is not
    explicitly provided, then neuropythy will either create a directory "HCP" in its default cache
    directory or, if there is no default cache directory, it will create a temporary directory just
    for the current Python session (this is deleted when Python exits).

    The ny.data['hcp'] object itself is a neuropythy Dataset object; the only relevant data field
    this dataset object contains is the subjects member, which is a lazy-map of the  HCP_1200
    subjects; e.g., ny.data['hcp'].subjects[100610] will yield the subject object for HCP subject
    100610.
    '''
    def __init__(self,
                 credentials=Ellipsis, release=Ellipsis, database=Ellipsis,
                 metadata_path=None, genetic_path=None, behavioral_path=None,
                 default_alignment=Ellipsis, cache_directory=Ellipsis,
                 meta_data=None, create_directories=True, create_mode=0o755):
        cdir = cache_directory
        if cdir is Ellipsis: cdir = config['hcp_auto_path']
        if cdir is Ellipsis:
            cdir = config['hcp_subject_paths']
            if cdir is not None: cdir = next(iter(cdir), None)
        HCPMetaDataset.__init__(self, 'hcp',
                                meta_data=meta_data,
                                cache_directory=cdir,
                                create_directories=create_directories,
                                create_mode=create_mode,
                                cache_required=True,
                                metadata_path=metadata_path,
                                genetic_path=genetic_path,
                                behavioral_path=behavioral_path)
        self.release = release
        self.database = database
        self.credentials = credentials
        self.default_alignment = default_alignment
    @pimms.param
    def default_alignment(df):
        '''
        hcp.default_alignment is the default alignment used by the HCP dataset. By default this is
        'MSMAll'.
        '''
        if df is None or df is Ellipsis: df = config['hcp_auto_default_alignment']
        df = df.lower()
        if   df in ['all', 'msmall', 'auto', 'automatic']: return 'MSMAll'
        elif df in ['sulc', 'msmsulc']: return 'MSMSulc'
        else: raise ValueError('default_alignment must be MSMAll or MSMSulc')
    @pimms.param
    def release(r):
        '''
        hcp.release is the release name for the HCP dataset object; by default this is 'HCP_1200' or
        whatever value is configured in neuropythy.config.
        '''
        if r is Ellipsis or r is None: return config['hcp_auto_release']
        elif not pimms.is_str(r): raise ValueError('HCP release must be a string')
        else: return r
    @pimms.param
    def database(r):
        '''
        hcp.database is the database name for the HCP dataset object; by default this is
        'hcp-openaccess' or whatever value is configured in neuropythy.config.
        '''
        if r is Ellipsis or r is None: return config['hcp_auto_database']
        elif not pimms.is_str(r): raise ValueError('HCP database must be a string')
        else: return r
    @pimms.param
    def credentials(c):
        '''
        hcp.credentials is the credential pair used by the HCP Dataset object hcp.
        '''
        if c is Ellipsis or c is None: return config['hcp_credentials']
        else: return to_credentials(c)
    @pimms.value
    def url(database, release):
        '''
        hcp.url is the base url of the HCP S3 subjects path; this is typically
        's3://hcp-openaccess/HCP_1200'.
        '''
        return 's3://' + database + '/' + release
    @pimms.value
    def s3_path(url, credentials, cache_directory):
        '''
        hcp.s3_path is the pseudo_path object that is used to create the various HCP subjects.
        '''
        return pseudo_path(url, cache_path=cache_directory, delete=False, credentials=credentials)
    @pimms.value
    def s3fs(s3_path):
        '''
        hcp.s3fs is the s3fs object that maintains a connection to the amazon s3.
        '''
        return s3_path._path_data['pathmod'].s3fs
    @staticmethod
    def _load_subject(ppath, url, credentials, cache_directory, df, sid):
        '''
        Creates a pseudo-path for the given subject and loads that subject as an hcp subject then
        returns it.
        '''
        from neuropythy.hcp import subject
        from posixpath import join as urljoin
        pm    = ppath._path_data['pathmod']
        creds = ppath.credentials
        # Make the pseudo-path; if auto-downloading is set to false, we want to create a pseudo-path
        # from only the cache path
        if config['hcp_auto_download'] in [False, None]:
            ppath = pseudo_path(os.path.join(cache_directory, str(sid)), delete=False)
        else:
            ppath = pseudo_path((pm.s3fs, urljoin(url, str(sid))),
                                credentials=credentials,
                                cache_path=os.path.join(cache_directory, str(sid)),
                                delete=False) # the base dir will be deleted if needs be
        return subject(ppath, default_alignment=df, name=str(sid))
    @pimms.value
    def _subjects(url, s3_path, default_alignment, cache_directory):
        '''
        hcp._subjects is a lazy persistent map of all the subjects that are part of the HCP_1200
        dataset. Unlike the hcp.subjects member, _subjects do not contain retinotopic mapping data
        for those subjects with available retinotopies from the ny.data['hcp_retinotopy'] dataset.
        '''
        # we get the subject list by doing an ls of the amazon bucket, but we fallback to a constant
        # list (in neuropythy.hcp)
        logging.info('HCPDataset: Getting HCP subject list...')
        aws_ips = ['99.84.127.26', '99.84.127.10', '99.84.127.3', '99.84.127.119']
        try:
            # this will raise an error if the internet is not alive
            dummy = urllib.request.urlopen('http://216.58.192.142', timeout=1)
            # internet is alive, so continue...
            pm = s3_path._path_data['pathmod']
            subdirs = []
            for s in pm.s3fs.ls(url):
                s = pm.split(s)[-1]
                try: s = int(s)
                except Exception: continue
                subdirs.append(s)
        except Exception:
            logging.info('       ...  Failed; falling back on builtin list.')
            subdirs = hcp.subject_ids
        creds = s3_path.credentials
        # okay, we will make a loader for each of these:
        ss = {sid: curry(HCPDataset._load_subject,
                         s3_path, url, creds, cache_directory, default_alignment, sid)
              for sid in subdirs}
        return pimms.lazy_map(ss)
    @pimms.value
    def subject_ids(_subjects):
        '''
        hcp.subject_ids is a tuple of the subject ids found in the HCP dataset.
        '''
        return tuple(_subjects.keys())
    @pimms.value
    def subjects(_subjects):
        '''
        hcp.subjects is a lazy persistent map of all the subjects that are part of the HCP_1200
        dataset. Subjects with valid retinotopic mapping data (assuming that the
        ny.data['hcp_retinotopy'] dataset has been initialized) include retinotopic mapping data
        as part of their property data.
        '''
        try:
            from neuropythy import data
            dset = data['hcp_retinotopy']
            subs = dset.subjects
        except Exception: return _subjects
        # okay, so far so good; let's setup the subject updating function:
        sids = set(list(_subjects.keys()))
        def _add_retino(sid):
            if sid in subs: return subs[sid]
            else:           return _subjects[sid]
        return pimms.lazy_map({sid: curry(_add_retino, sid) for sid in six.iterkeys(_subjects)})
    def download(self, sid):
        '''
        ny.data['hcp'].download(sid) downloads all the data understood by neuropythy for the given
        HCP subject id; the data are downloaded from the Amazon S3 into the path given by the 
        'hcp_auto_path' config item then returns a list of the downloaded files.
        '''
        # we can do this in quite a sneaky way: get the subject, get their filemap, force all the
        # paths in the subject to be downloaded using the pseudo-path, return the cache path!
        sub   = self.subjects[sid]
        fmap  = sub.meta_data['file_map']
        ppath = fmap.path
        fls   = []
        logging.info('Downloading HCP subject %s structure data...' % (sid,))
        for fl in six.iterkeys(fmap.data_files):
            logging.info('  * Downloading file %s for subject %s' % (fl, sid))
            try:
                fls.append(ppath.local_path(fl))
            except ValueError as e:
                if len(e.args) != 1 or not e.args[0].startswith('getpath:'): raise
                else: logging.info('    (File %s not found for subject %s)' % (fl, sid))
        logging.info('Subject %s donwnload complete!' % (sid,))
        return fls
# we wrap this in a lambda so that it gets loaded when requested (in case the config changes between
# when this gets run and when the dataset gets requested)
add_dataset('hcp', lambda:HCPDataset().persist())

# HCPRetinotopyDataset #############################################################################
def to_retinotopy_cache_path(p):
    '''
    to_retinotopy_cache_path(p) yields p if p is a directory and raises an exception otherwise.
    '''
    if pimms.is_str(p) and os.path.isdir(p): return os.path.normpath(p)
    elif p is Ellipsis: return p
    else: return None
config.declare('hcp_retinotopy_cache_path', environ_name='HCP_RETINOTOPY_CACHE_PATH',
               filter=to_retinotopy_cache_path, default_value=Ellipsis)
def _prefix_keys(m, prefix):
    '''
    Lazily adds the given prefix to all keys in m, which must all be strings, and returns the new
    lazy map.
    '''
    f = lambda k:m[k]
    return pimms.lazy_map({(prefix+k): curry(f, k) for k in six.iterkeys(m)})    
hcp_retinotopy_property_names = ('polar_angle', 'eccentricity', 'radius', 'variance_explained',
                                 'mean_signal', 'gain', 'x', 'y')
hcp_retinotopy_property_files = pyr.pmap({'polar_angle':'angle', 'eccentricity':'eccen', 
                                          'mean_signal':'means', 'gain':'const', 'radius':'prfsz',
                                          'variance_explained':'vexpl', 'x':'xcrds', 'y':'ycrds'})
@pimms.calc(*hcp_retinotopy_property_names)
def calc_native_properties(native_hemi, fs_LR_hemi, prefix, resolution, subject_id=None,
                           method='nearest', cache_directory=None, alignment='MSMAll'):
    '''
    calc_native_properties is a pimms calculator that requires a native hemisphere, an fs_LR
    hemisphere, and a property prefix and yields the interpolated retinotopic mapping properties.
    '''
    method = to_interpolation_method(method)
    # first: check for a cache file:
    if cache_directory is not None:
        if subject_id is None: pth = cache_directory
        else: pth = os.path.join(cache_directory, str(subject_id), 'retinotopy')
        # we load all or none here:
        ftr = hcp_retinotopy_property_files
        mtd = '.' if method == 'nearest' else '.linear.'
        flp = '%s.%s%%s.%s%snative%dk.mgz' % (native_hemi.chirality,prefix,alignment,mtd,resolution)
        fls = {k:os.path.join(pth, flp % ftr[k]) for k in hcp_retinotopy_property_names}
        try: return {k:nyio.load(v, 'mgh') for (k,v) in six.iteritems(fls)}
        except Exception: pass
    else: fls = None

    # make sure we can grab one of the properties before we print anything...
    ptest = fs_LR_hemi.prop(prefix+hcp_retinotopy_property_names[0])
    # if that didn't fail, print the message...
    logging.info('HCP Dataset: Interpolating retinotopy for HCP subject %s / %s (method: %s)...'
                 % (subject_id, native_hemi.chirality, method))
    p = fs_LR_hemi.interpolate(native_hemi,
                               {k:(prefix+k) for k in hcp_retinotopy_property_names
                                if k not in ['polar_angle', 'eccentricity']},
                               method=method)
    # calculate the angle/eccen from the x and y values
    theta = np.arctan2(p['y'], p['x'])
    p = pimms.assoc(p,
                    polar_angle=np.mod(90 - 180/np.pi*theta + 180, 360) - 180,
                    eccentricity=np.hypot(p['x'], p['y']))
    # write cache and return
    if fls is not None:
        try:
            for (k,v) in six.iteritems(p): nyio.save(fls[k], v)
        except Exception as e:
            tup = (subject_id, native_hemi.chirality, type(e).__name__ + str(e.args))
            warnings.warn('cache write failed for HCP retinotopy subject %s / %s: %s' % tup,
                          RuntimeWarning)
    return p
interpolate_native_properties = pimms.plan({'native_properties': calc_native_properties})

@pimms.immutable
class HCPRetinotopyDataset(HCPMetaDataset):
    '''
    neuropythy.data['hcp_retinotopy'] is a Dataset containing the publicly provided data from the
    Benson et al. (2018; DOI:10.1167/18.13.23) paper on the HCP 7T retinotopy dataset. For more
    information see the paper's OSF site (https://osf.io/bw9ec/)

    You do not have to explicitly configure anything in order to use this dataset: the default
    behavior is to use the the same cache directory as the 'hcp' dataset, meaning that your subject
    directories will gain 'retinotopy' subdirectories containing cached mgz files. These
    'retinotopy' subdirectories exist at the same level as 'MNINonLinear' and 'T1w'.

    It is recommended that you interact with the data via the neuropythy.data['hcp'] dataset--this
    dataset stores the structural data necessary to make sense of the retinotopic data and will
    automatically include the relevant retinotopic data as properties attached to any subject
    requested that has retinotopic data.

    If you wish to explicitly disable hcp_retinotopy downloading, you can do so by setting the
    'hcp_auto_download' config item to either False or 'structure' (indicating that structural
    downloads should continue but retinotopic ones should not). Note that this latter setting does
    not actually prevent you from inducing downloading via the 'hcp_retinotopy' dataset directly,
    but it will prevent the 'hcp' dataset from inducing such downloads.

    The default behavior of the hcp_retinotopy dataset is to put cache files in the same directories
    as the auto-downloaded HCP subject data; a separate cache directory can be provided via the
    neuropythy hcp_retinotopy_cache_path config item.
    '''
    default_url = 'osf://bw9ec/'
    retinotopy_files = pyr.pmap({32:'prfresults.mat', 59:'prfresults59k.mat'})
    retinotopy_prefix = 'prf'
    lowres_retinotopy_prefix = 'lowres-prf'
    highres_retinotopy_prefix = 'highres-prf'
    subject_ids = tuple([100610, 102311, 102816, 104416, 105923, 108323, 109123, 111312,
                         111514, 114823, 115017, 115825, 116726, 118225, 125525, 126426,
                         128935, 130114, 130518, 131217, 131722, 132118, 134627, 134829,
                         135124, 137128, 140117, 144226, 145834, 146129, 146432, 146735,
                         146937, 148133, 150423, 155938, 156334, 157336, 158035, 158136,
                         159239, 162935, 164131, 164636, 165436, 167036, 167440, 169040,
                         169343, 169444, 169747, 171633, 172130, 173334, 175237, 176542,
                         177140, 177645, 177746, 178142, 178243, 178647, 180533, 181232,
                         181636, 182436, 182739, 185442, 186949, 187345, 191033, 191336,
                         191841, 192439, 192641, 193845, 195041, 196144, 197348, 198653,
                         199655, 200210, 200311, 200614, 201515, 203418, 204521, 205220,
                         209228, 212419, 214019, 214524, 221319, 233326, 239136, 246133,
                         249947, 251833, 257845, 263436, 283543, 318637, 320826, 330324,
                         346137, 352738, 360030, 365343, 380036, 381038, 385046, 389357,
                         393247, 395756, 397760, 401422, 406836, 412528, 429040, 436845,
                         463040, 467351, 525541, 536647, 541943, 547046, 550439, 552241,
                         562345, 572045, 573249, 581450, 585256, 601127, 617748, 627549,
                         638049, 644246, 654552, 671855, 680957, 690152, 706040, 724446,
                         725751, 732243, 751550, 757764, 765864, 770352, 771354, 782561,
                         783462, 789373, 814649, 818859, 825048, 826353, 833249, 859671,
                         861456, 871762, 872764, 878776, 878877, 898176, 899885, 901139,
                         901442, 905147, 910241, 926862, 927359, 942658, 943862, 951457,
                         958976, 966975, 971160, 973770, 995174, 999997, 999998, 999999])

    # these expect value % (hemi, alignment, surfacename)
    _retinotopy_cache_tr = pimms.persist({
        # the native filename's here aren't actually used; they would also include a '.linear.' for
        # linear interpolation right before the 'native59k' or 'native32k' (these are the filenames
        # for nearest interpolation)
        'native': { 
            (highres_retinotopy_prefix + '_polar_angle')        :'%s.split%s_angle.%s.native59k.mgz',
            (highres_retinotopy_prefix + '_eccentricity')       :'%s.split%s_eccen.%s.native59k.mgz',
            (highres_retinotopy_prefix + '_radius')             :'%s.split%s_prfsz.%s.native59k.mgz',
            (highres_retinotopy_prefix + '_variance_explained') :'%s.split%s_vexpl.%s.native59k.mgz',
            (highres_retinotopy_prefix + '_mean_signal')        :'%s.split%s_means.%s.native59k.mgz',
            (highres_retinotopy_prefix + '_gain')               :'%s.split%s_const.%s.native59k.mgz',
            (lowres_retinotopy_prefix + '_polar_angle')         :'%s.split%s_angle.%s.native32k.mgz',
            (lowres_retinotopy_prefix + '_eccentricity')        :'%s.split%s_eccen.%s.native32k.mgz',
            (lowres_retinotopy_prefix + '_radius')              :'%s.split%s_prfsz.%s.native32k.mgz',
            (lowres_retinotopy_prefix + '_variance_explained')  :'%s.split%s_vexpl.%s.native32k.mgz',
            (lowres_retinotopy_prefix + '_mean_signal')         :'%s.split%s_means.%s.native32k.mgz',
            (lowres_retinotopy_prefix + '_gain')                :'%s.split%s_const.%s.native32k.mgz'},
        'LR32k': {                                             
            (lowres_retinotopy_prefix + '_polar_angle')         :'%s.split%s_angle.32k.mgz',
            (lowres_retinotopy_prefix + '_eccentricity')        :'%s.split%s_eccen.32k.mgz',
            (lowres_retinotopy_prefix + '_radius')              :'%s.split%s_prfsz.32k.mgz',
            (lowres_retinotopy_prefix + '_variance_explained')  :'%s.split%s_vexpl.32k.mgz',
            (lowres_retinotopy_prefix + '_mean_signal')         :'%s.split%s_means.32k.mgz',
            (lowres_retinotopy_prefix + '_gain')                :'%s.split%s_const.32k.mgz'},
        'LR59k': {                                             
            (highres_retinotopy_prefix + '_polar_angle')        :'%s.split%s_angle.59k.mgz',
            (highres_retinotopy_prefix + '_eccentricity')       :'%s.split%s_eccen.59k.mgz',
            (highres_retinotopy_prefix + '_radius')             :'%s.split%s_prfsz.59k.mgz',
            (highres_retinotopy_prefix + '_variance_explained') :'%s.split%s_vexpl.59k.mgz',
            (highres_retinotopy_prefix + '_mean_signal')        :'%s.split%s_means.59k.mgz',
            (highres_retinotopy_prefix + '_gain')               :'%s.split%s_const.59k.mgz'}})
    
    def __init__(self, url=Ellipsis, cache_directory=Ellipsis, interpolation_method=Ellipsis,
                 metadata_path=None, genetic_path=None, behavioral_path=None,
                 meta_data=None, create_directories=True, create_mode=0o755):
        cdir = cache_directory
        if cdir is Ellipsis: cdir = config['hcp_auto_path']
        if cdir is Ellipsis:
            cdir = config['hcp_subject_paths']
            if cdir is not None: cdir = next(iter(cdir), None)
        HCPMetaDataset.__init__(self, 'hcp_retinotopy',
                                meta_data=meta_data,
                                cache_directory=cdir,
                                create_directories=create_directories,
                                create_mode=create_mode,
                                cache_required=True,
                                metadata_path=metadata_path,
                                genetic_path=genetic_path,
                                behavioral_path=behavioral_path)
        if url is Ellipsis: url = HCPRetinotopyDataset.default_url
        self.url = url
        self.interpolation_method = interpolation_method
    @pimms.param
    def url(u):
        '''
        ny.data['hcp_retinotopy'].url is the url from which the retinotopy data is loaded.
        '''
        if not pimms.is_str(u): raise ValueError('url must be a string')
        return u
    @pimms.param
    def interpolation_method(im):
        '''
        ny.data['hcp_retinotopy'].interpolation_method is a string, either 'nearest' (default) or
        'linear', which specifies whether nearest or linear interpolation should be used when
        interpolating retinotopy data from the fs_LR meshes onto the native meshes.
        '''
        if im is Ellipsis or im is None: return config['hcp_retinotopy_interpolation_method']
        else: return to_interpolation_method(im)
    @pimms.value
    def pseudo_path(url, cache_directory):
        '''
        ny.data['hcp_retinotopy'].pseudo_path is the psueod-path object responsible for loading the
        retinotopy data.
        '''
        return pseudo_path(url, cache_path=cache_directory).persist()
    @pimms.value
    def cifti_data(pseudo_path):
        '''
        ny.data['hcp_retinotopy'].cifti_data is a tuple of lazy maps of the 32k and 59k data arrays,
        reorganized into 'visual' retinotopic coordinates. The tuple elements represent the
        (full, split1, split2) solutions.
        '''
        # our loader function:
        def _load(res, split):
            import h5py
            flnm = HCPRetinotopyDataset.retinotopy_files[res]
            logging.info('HCPRetinotopyDataset: Loading split %d from file %s...' % (split, flnm))
            # Before we get this from the 
            flnm = pseudo_path.local_path(flnm)
            with h5py.File(flnm, 'r') as f:
                sids = np.array(f['subjectids'][0], dtype='int')
                data = np.array(f['allresults'][split])
            sids.setflags(write=False)
            # convert these into something more coherent
            tmp = hcp.cifti_split(data)
            for q in tmp: q.setflags(write=False)
            return pyr.pmap(
                {h: pyr.m(prf_polar_angle        = np.mod(90 - dat[:,0] + 180, 360) - 180,
                          prf_eccentricity       = dat[:,1],
                          prf_radius             = dat[:,5],
                          prf_variance_explained = dat[:,4]/100.0,
                          prf_mean_signal        = dat[:,3],
                          prf_gain               = dat[:,2],
                          prf_x                  = dat[:,1]*np.cos(np.pi/180*dat[:,0]),
                          prf_y                  = dat[:,1]*np.sin(np.pi/180*dat[:,0]))
                 for (h,dat) in zip(['lh','rh','subcortical'], tmp)})
        splits = [pimms.lazy_map({res: curry(_load, res, split)
                                  for res in six.iterkeys(HCPRetinotopyDataset.retinotopy_files)})
                  for split in [0,1,2]]
        return tuple(splits)
    @pimms.value
    def subject_order(pseudo_path):
        '''
        subject_order is a mapping of subject ids to the offset at which they appear in the cifti
        data.
        '''
        import h5py
        def _load(res):
            flnm = HCPRetinotopyDataset.retinotopy_files[res]
            flnm = pseudo_path.local_path(flnm)
            logging.info('HCPRetinotopyDataset: Loading subjects from file %s...' % flnm)
            with h5py.File(flnm, 'r') as f:
                sids = np.array(f['subjectids'][0], dtype='int')
            return pyr.pmap({sid:ii for (ii,sid) in enumerate(sids)})
        return pimms.lazy_map({res: curry(_load, res)
                               for res in six.iterkeys(HCPRetinotopyDataset.retinotopy_files)})
    @pimms.value
    def retinotopy_data(cifti_data, subject_order, cache_directory,create_directories,create_mode):
        '''
        ny.data['hcp_retinotopy'].retinotopy_data is a nested-map data structure representing the
        retinotopic mapping data for each subject. The first layer of keys in retinotopy_data is
        the subject id, including the 99999* average subjects. The second layer of keys is the name
        of the hemisphere, and the third layer of keys are property names (the final layer is always
        constructed of lazy maps).
        '''
        rpfx = HCPRetinotopyDataset.retinotopy_prefix
        sids = HCPRetinotopyDataset.subject_ids
        # how we load data:
        def _load_LR(split, sid, h, res, prop, flpatt):
            pth = os.path.join(cache_directory, str(sid), 'retinotopy')
            if not os.path.exists(pth) and create_directories:
                try:
                    os.makedirs(pth, create_mode)
                except Exception as e:
                    w = "an exception of type %s was raised while trying to make cache directory %s"
                    warnings.warn(w % (type(e), pth))
            flnm = os.path.join(pth, flpatt % (h,split))
            # deduce the generic property name from the given prop name
            pp = rpfx + prop.split(rpfx)[1]
            # as a general rule: we don't want to touch cifti_split lazy-maps because they force the
            # entire .mat file to load... BUT if we have already loaded it, we're better off sharing
            # the loaded data instead of reloading cache
            # parse the prop -- we can remove the front of it
            cdat = cifti_data[split]
            okwrite = True
            sii = subject_order[res][sid]
            if not cdat.is_lazy(res) or not os.path.isfile(flnm): dat = cdat[res][h][pp][sii]
            else:
                try: return nyio.load(flnm)
                except Exception:
                    warnings.warn('failed to load HCP retinotopy cache file %s'%flnm)
                    okwrite = False
                dat = cdat[res][h][pp][sii]
            if okwrite:
                try: nyio.save(flnm, dat)
                except Exception as e:
                    warnings.warn("Error when trying to save cache file: %s" % (type(e),))
            return dat
        # okay, the base properties for the 32k and 59k meshes:
        trs = HCPRetinotopyDataset._retinotopy_cache_tr
        base_props = {
            sid: {h: {split: {res: pimms.lazy_map({p: curry(_load_LR, split, sid, h, res, p, flp)
                                                   for (p,flp) in six.iteritems(trs['LR%dk'%res])})
                              for res in [32, 59]}
                      for split in [0,1,2]}
                  for h in ['lh','rh']}
            for sid in sids}
        # okay, now we build on top of these: we add in the x/y properties
        def _add_xy(dat):
            k = next(six.iterkeys(dat))
            prefix = k.split('_')[0] + '_'
            (ang,ecc) = [dat[prefix + k] for k in ('polar_angle', 'eccentricity')]
            tht = np.pi/180 * (90 - ang)
            (x,y) = [ecc*np.cos(tht), ecc*np.sin(tht)]
            return pimms.assoc(dat, prefix + 'x', x, prefix + 'y', y)
        xy_props = {
            sid: {h: {split: pimms.lazy_map({res: curry(_add_xy, rdat)
                                             for (res,rdat) in six.iteritems(spdat)})
                      for (split,spdat) in six.iteritems(hdat)}
                  for (h,hdat) in six.iteritems(sdat)}
            for (sid,sdat) in six.iteritems(base_props)}
        # okay, that's it; just organize it into the desired shape
        def _reorg(hdat, res, ks):
            pfx = ks[0].split('_')[0] + '_'
            ks = ks + [pfx + 'x', pfx + 'y']
            f = lambda s,k:s[res][k]
            return pimms.lazy_map(
                {(pre+k): curry(f, s, k)
                 for k in ks
                 for (s,pre) in zip([hdat[0],hdat[1],hdat[2]], ['', 'split1-', 'split2-'])})
        r = {sid: {('%s_LR%dk'%(h,res)): _reorg(hdat, res, list(base_props[sid][h][0][res].keys()))
                   for (h,hdat) in six.iteritems(sdat)
                   for res in [32,59]}
             for (sid,sdat) in six.iteritems(xy_props)}
        return pimms.persist(r)
    @pimms.value
    def subjects(retinotopy_data, cache_directory, create_mode, create_directories,
                 interpolation_method):
        '''
        hcp_retinotopy.subjects is a lazy map whose keys are HCP subject IDs (of those subjects with
        valid retinotopic mapping data) and whose values are subject objects (obtained from the
        ny.data['hcp']._subjects map) with the addition of retinotopic mapping properties.
        '''
        alltrs = HCPRetinotopyDataset._retinotopy_cache_tr
        lrtrs = {32: alltrs['LR32k'], 59: alltrs['LR59k']}
        lpfx = HCPRetinotopyDataset.lowres_retinotopy_prefix
        hpfx = HCPRetinotopyDataset.highres_retinotopy_prefix
        nttrs = {32: {k:v for (k,v) in six.iteritems(alltrs['native']) if k.startswith(lpfx)},
                 59: {k:v for (k,v) in six.iteritems(alltrs['native']) if k.startswith(hpfx)}}
        def _prep(sid):
            # see if we can get the subject from the 'hcp' dataset:
            try:
                from neuropythy import data
                hcp = data['hcp']
                sub = hcp._subjects[sid]
            except Exception: hcp = None
            if sid > 999990: raise ValueError("IDs 999997-999999 are not real HCP subjects")
            elif hcp is None: raise ValueError('could not load subject object for sid %s' % (sid,))
            # okay, we need to prep this subject; the initial part is easy: copy the properties
            # for the fs_LR meshes over to them:
            hems = hems0 = sub.hemis
            def _add_LR(hnm, res):
                hem = hems0[hnm]
                return hem.with_prop(retinotopy_data[sid]['%s_LR%dk' % (hnm[:2], res)])
            for h in ['lh','rh']:
                for res in [32,59]:
                    for align in ['MSMAll', 'MSMSulc']:
                        hnm = '%s_LR%dk_%s' % (h,res,align)
                        hems = hems.set(hnm, curry(_add_LR, hnm, res))
            # okay, we have the data transferred over the the fs_LR hemispheres now; we just need
            # to do some interpolation for the native hemispheres
            hems1 = hems
            def _get(inp, k): return inp[k]
            def _interp_hem(hemi, h, res, align):
                pfls = nttrs[res]
                pfx = next(six.iterkeys(pfls)).split('_')[0] + '_'
                inpargs = dict(native_hemi=hemi, fs_LR_hemi=hems1['%s_LR%dk_%s' % (h, res, align)],
                               subject_id=sid,   method=interpolation_method,  resolution=res,
                               alignment=align,  cache_directory=cache_directory)
                lm = pimms.lazy_map({(p+k): curry(_get, inp, k)
                                     for spl in ['', 'split1-', 'split2-'] for p in [spl + pfx]
                                     for inp in [interpolate_native_properties(prefix=p, **inpargs)]
                                     for k   in hcp_retinotopy_property_names})
                return hemi.with_prop(lm)
            def _get_highest_res(hemi, k):
                try: x = hemi.prop(hpfx + '_' + k)
                except Exception as e: x = None
                if x is not None: return x
                try: return hemi.prop(lpfx + '_' +  k)
                except Exception as e: x = None
                if x is not None: return x
                raise ValueError('no retinotopy successfully loaded for hemi', hemi)
            def _interp_nat(h, align):
                # okay, let's get the hemisphere we are modifying...
                hemi = hem0 = hems1['%s_native_%s' % (h, align)]
                # we're going to interp from both the 32k and 59k meshes (if possible)
                for res in [32, 59]:
                    # These may fail due to not having files for one or the other (if not loading
                    # from S3); this is okay, though.
                    try: hemi = _interp_hem(hemi, h, res, align)
                    except Exception: pass
                if hemi is hem0:
                    warnings.warn('unable to interpolate retinotopy for any hemisphere '
                                  'of hcp_retinotopy subject')
                # add the 'best/standard' prf header:
                pfx = HCPRetinotopyDataset.retinotopy_prefix
                lm = pimms.lazy_map({(pfx + '_' + k): curry(_get_highest_res, hemi, k)
                                     for k in hcp_retinotopy_property_names})
                # return with these properties:
                return hemi.with_prop(lm)
            for h in ['lh','rh']:
                for align in ['MSMAll','MSMSulc']:
                    hems = hems.set(h + '_native_' + align, curry(_interp_nat, h, align))
            # fix the hemisphere aliases based on default alignment:
            default_alignment = sub.meta_data.get('default_alignment',
                                                  config['hcp_default_alignment'])
            hems2 = hems
            for h in ['lh_native',  'rh_native',  'lh_LR32k',   'rh_LR32k',
                      'lh_LR59k',   'rh_LR59k',   'lh_LR164k',  'rh_LR164k']:
                hems = hems.set(h, curry(lambda h:hems2[h+'_'+default_alignment], h))
            hems = hems.set('lh', lambda:hems2['lh_native_' + default_alignment])
            hems = hems.set('rh', lambda:hems2['rh_native_' + default_alignment])
            return sub.copy(hemis=hems)
        # we just need to call down to this prep function lazily:
        return pimms.lazy_map({sid: curry(_prep, sid) for sid in six.iterkeys(retinotopy_data)})
add_dataset('hcp_retinotopy', lambda:HCPRetinotopyDataset().persist())



####################################################################################################
# neuropythy/util/__init__.py
# This file defines the general tools that are available as part of neuropythy.

from .core     import (ObjectWithMetaData, normalize, denormalize,
                       to_hemi_str, to_affine, simplex_averaging_matrix, simplex_summation_matrix,
                       is_dataframe, to_dataframe, dataframe_select, dataframe_except,
                       is_image, is_image_header, curry,
                       numel, rows, part, hstack, vstack, repmat, replace_close, chop,
                       flatter, flattest, is_tuple, is_list, is_set,
                       plus, cplus, minus, cminus, times, ctimes, 
                       inv, zinv, divide, cdivide, zdivide, czdivide, power, cpower, inner,
                       sine, cosine, tangent, cotangent, secant, cosecant,
                       arcsine, arccosine, arctangent,
                       naneq, nanne, nanlt, nanle, nangt, nange, nanlog,
                       library_path, address_data, is_address, address_interpolate,
                       AutoDict, auto_dict, curve_spline, curve_intersection, close_curves,
                       is_curve_spline, to_curve_spline, CurveSpline,
                       DataStruct, data_struct, tmpdir, dirpath_to_list, apply_affine, try_until)
from .conf     import (config, to_credentials, detect_credentials, load_credentials)
from .filemap  import (FileMap, file_map, is_file_map, pseudo_path, is_pseudo_path, to_pseudo_path,
                       osf_crawl, url_download)
from .labels   import (label_colors, is_label_index, label_index, label_indices, to_label_index)




####################################################################################################
# neuropythy/freesurfer/core.py
# Simple tools for use with FreeSurfer in Python
# By Noah C. Benson

import numpy                        as np
import nibabel                      as nib
import nibabel.freesurfer.io        as fsio
import nibabel.freesurfer.mghformat as fsmgh
import pyrsistent                   as pyr
from   six.moves                import collections_abc as collections
import os, warnings, six, pimms

from .. import geometry as geo
from .. import mri      as mri
from .. import io       as nyio
#import ..io as nyio

from ..util import (config, library_path, curry, pseudo_path, is_pseudo_path, data_struct,
                    label_indices, label_index, to_label_index, is_tuple, file_map, to_pseudo_path,
                    try_until)

def library_freesurfer_subjects():
    '''
    library_freesurfer_subjects() yields the path of the neuropythy library's FreeSurfer subjects
      directory.
    '''
    return os.path.join(library_path(), 'data')

####################################################################################################
# FreeSurfer home and where to find FreeSurfer LUTs and such
# some functions to load in the color luts:
@nyio.importer('FreeSurferLUT', 'LUT.txt')
def load_LUT(filename, to='data'):
    '''
    load_LUT(filename) loads the given filename as a FreeSurfer LUT.

    The optional argument to (default: 'data') specifies how the LUT should be interpreted; it can
    be any of the following:
      * 'data' specifies that a dataframe should be returned.
    '''
    from neuropythy.util import to_dataframe
    import pandas
    # start by slurping in the text:
    dat = pandas.read_csv(filename, comment='#', sep='\s+', names=['id', 'name', 'r','g','b','a'])
    # if all the alpha values are 0, we set them to 1 (not sure why freesurfer does this)
    dat['a'] = 255 - dat['a']
    if pimms.is_str(to): to = to.lower()
    if   to is None: return dat
    elif to == 'data':
        df = to_dataframe({'id': dat['id'].values, 'name': dat['name'].values})
        df['color'] = dat.apply(lambda r: [r[k]/255.0 for k in ['r','g','b','a']], axis=1)
        df.set_index('id', inplace=True)
        return df
    else: raise ValueError('Unknown to argument: %s' % to)
# A function to load in default data from the freesurfer home: e.g., the default LUTs
def _load_fshome(path):
    luts = {'aseg': 'ASegStatsLUT.txt',
            'wm':   'WMParcStatsLUT.txt',
            'all':  'FreeSurferColorLUT.txt'}
    luts = {k: curry(load_LUT, os.path.join(path, v)) for (k,v) in six.iteritems(luts)}
    luts = pimms.lazy_map(luts)
    # put these into the label indices
    global label_indices
    label_indices['freesurfer']      = label_index(luts['all'])
    label_indices['freesurfer_aseg'] = label_index(luts['aseg'])
    label_indices['freesurfer_wm']   = label_index(luts['wm'])
    return data_struct(luts=luts, path=path)
config.declare_dir('freesurfer_home', environ_name='FREESURFER_HOME', filter=_load_fshome)

####################################################################################################
# Subject Directory and where to find Subjects
def to_subject_paths(paths, previous_paths=None):
    '''
    to_subject_paths(paths) accepts either a string that is a :-separated list of directories or a
      list of directories and yields a list of all the existing directories.
    '''
    if paths is None: return []
    if pimms.is_str(paths): paths = paths.split(':')
    paths = [os.path.expanduser(p) for p in paths]
    if previous_paths is not None: paths = previous_paths + paths
    return [p for p in paths if os.path.isdir(p)]
def freesurfer_paths_merge(p0, p1):
    '''
    freesurfer_paths_merge(p0, p1) yields to_subject_paths(p0) + to_subject_paths(p1).
    '''
    return to_subject_paths(p0) + to_subject_paths(p1)
config.declare('freesurfer_subject_paths', environ_name='SUBJECTS_DIR', filter=to_subject_paths,
               merge=freesurfer_paths_merge)

def subject_paths():
    '''
    subject_paths() yields a list of paths to Freesurfer subject directories in which subjects are
      automatically searched for when identified by subject-name only. These paths are searched in
      the order returned from this function.

    If you must edit these paths, it is recommended to use add_subject_path, and clear_subject_paths
    functions.
    '''
    return config['freesurfer_subject_paths']

def clear_subject_paths(subpaths):
    '''
    clear_subject_paths() resets the freesurfer subject paths to be empty and yields the previous
      list of subject paths.
    '''
    sd = config['freesurfer_subject_paths']
    config['freesurfer_subject_paths'] = []
    return sd

def add_subject_path(path, index=0):
    '''
    add_subject_path(path) will add the given path to the list of subject directories in which to
      search for Freesurfer subjects. The optional argument index may be given to specify the
      precedence of this path when searching for a new subject; the default, 0, always inserts the
      path at the front of the list; a value of k indicates that the new list should have the new
      path at index k.
    The path may contain :'s, in which case the individual directories are separated and added.
    If the given path is not a directory or the path could not be inserted, yields False;
    otherwise, yields True. If the string contains a : and multiple paths, then True is yielded
    only if all paths were successfully inserted.
    See also subject_paths.
    '''
    paths = [p for p in path.split(':') if len(p) > 0]
    if len(paths) > 1:
        tests = [add_subject_path(p, index=index) for p in reversed(paths)]
        return all(t for t in tests)
    else:
        fsp = config['freesurfer_subject_paths']
        if not os.path.isdir(path): return False
        if path in fsp: return True
        try:
            if index is None or index is Ellipsis:
                sd = fsp + [path]
            else:
                sd = fsp + []
                sd.insert(index, path)
            config['freesurfer_subject_paths'] = sd
            return True
        except Exception:
            return False

if config['freesurfer_subject_paths'] is None:
    # If the subjects path wasn't provided, try a few common subject paths
    add_subject_path('/Applications/freesurfer/subjects')
    add_subject_path('/opt/freesurfer/subjects')
# Regardless, make sure we check the FreeSurfer home
if config['freesurfer_home'] is not None:
    add_subject_path(os.path.join(config['freesurfer_home'].path, 'subjects'), None)

def is_freesurfer_subject_path(path):
    '''
    is_freesurfer_subject_path(path) yields True if the given path appears to be a valid freesurfer
      subject path and False otherwise.
    is_freesurfer_subject_path(pdir) performs the same check for the given pseudo-dir object pdir.

    A path is considered to be freesurfer-subject-like if it contains the directories mri/, surf/,
    and label/. 
    '''
    needed = ['mri', 'surf', 'label']
    if   is_pseudo_path(path): return all(path.find(d) is not None for d in needed)
    elif os.path.isdir(path): return all(os.path.isdir(os.path.join(path, d)) for d in needed)
    else:                     return False
  
def find_subject_path(sub, check_path=True):
    '''
    find_subject_path(sub) yields the full path of a Freesurfer subject with the name given by the
      string sub, if such a subject can be found in the Freesurfer search paths. See also
      add_subject_path.

    If no subject is found, then None is returned.

    The optional argument check_path (default: True) may be set to False to indicate that sanity
    checks regarding the state of the file-system should be skipped. If check_path is set to True,
    then a directory will only be returned if it is a directory on the file-system and it contains
    at least the subdirectories mri/, label/, and surf/. If check_path is set to False, then
    find_subject_path will first search for any directory that satisfies the above requirements,
    and, failing that, will search for any directory matching sub that at least exists. If this also
    fails, then sub itself is returned (i.e., sub is presumed to be a valid path).
    '''
    sdirs = config['freesurfer_subject_paths']
    # if it's a full/relative path already, use it:
    if ((not check_path or is_freesurfer_subject_path(sub)) and
        (check_path is None or os.path.isdir(sub))):
        return sub
    return next((p for sd in sdirs for p in [os.path.join(sd, sub)]
                 if ((not check_path or is_freesurfer_subject_path(p)) and
                     (check_path is None or os.path.isdir(p)))),
                None)

# Used to load immutable-like mgh objects
def _load_imm_mgh(flnm,u):
    img = fsmgh.load(flnm)
    try: img.dataobj.flags['WRITEABLE'] = False
    except Exception:
        d = np.asanyarray(img.dataobj)
        d.flags['WRITEABLE'] = False
        img = fsmgh.MGHImage(d, img.affine, img.header)
    return img

####################################################################################################
# Filemap
# This is the FreeSurfer file-map data, as neuropythy understands it.
def _load_label(flnm,u):
    (a,b) = nyio.load(flnm, 'freesurfer_label', to='all')
    a.setflags(write=False)
    b.setflags(write=False)
    return (a,b)
def _load_annot(flnm,u):
    (a,b) = nyio.load(flnm, 'freesurfer_annot', to='all')
    a.setflags(write=False)
    return (a,b.persist())
def _load_surfx(flnm,u):
    x = fsio.read_geometry(flnm)[0].T
    x.setflags(write=False)
    return x
def _load_prop(flnm,u):
    x = fsio.read_morph_data(flnm)
    x.setflags(write=False)
    return x
def _load_tris(flnm,u):
    x = fsio.read_geometry(flnm)[1]
    x.setflags(write=False)
    return x
def _load_geo(flnm,u):
    (x,t) = fsio.read_geometry(flnm)[:2]
    x = x.T
    x.setflags(write=False)
    t.setflags(write=False)
    return (x,t)    
freesurfer_subject_data_hierarchy = [['image'],                ['raw_image'],
                                     ['hemi', 'surface'],      ['hemi', 'tess'],
                                     ['hemi', 'registration'], ['hemi', 'property'],
                                     ['hemi', 'label'],        #['hemi', 'alt_label'],
                                     ['hemi', 'weight'],       ['hemi', 'alt_weight'],
                                     ['hemi', 'annot'],        ['hemi', 'alt_annot']]
freesurfer_subject_filemap_instructions = [
    'mri', [
        'rawavg.mgz',              ({'raw_image':'rawavg', 'load': _load_imm_mgh},
                                    {'image':'raw', 'load': _load_imm_mgh}),                     
        'orig.mgz',                ({'raw_image':'orig', 'load': _load_imm_mgh},
                                    {'image':'original', 'load': _load_imm_mgh}),                
        'orig_nu.mgz',             ({'raw_image':'orig_nu', 'load': _load_imm_mgh},
                                    {'image':'conformed', 'load': _load_imm_mgh}),
        'nu.mgz',                  ({'raw_image':'nu', 'load': _load_imm_mgh},
                                    {'image':'uniform', 'load': _load_imm_mgh}),
        'T1.mgz',                  ({'raw_image':'T1', 'load': _load_imm_mgh},
                                    {'image':'intensity_normalized', 'load': _load_imm_mgh}),
        'brainmask.auto.mgz',      ({'raw_image':'brainmask.auto', 'load': _load_imm_mgh},
                                    {'image':'auto_masked_brain', 'load': _load_imm_mgh}),
        'brainmask.mgz',           ({'raw_image':'brainmask', 'load': _load_imm_mgh},
                                    {'image':'masked_brain', 'load': _load_imm_mgh}),
        'norm.mgz',                ({'raw_image':'norm', 'load': _load_imm_mgh},
                                    {'image':'normalized', 'load': _load_imm_mgh}),
        'aseg.auto_noCCseg.mgz',   {'raw_image':'aseg.auto_noCCseg', 'load': _load_imm_mgh},
        'aseg.auto.mgz',           ({'raw_image':'aseg.auto', 'load': _load_imm_mgh},
                                    {'image':'auto_segmentation', 'load': _load_imm_mgh}),
        'aseg.presurf.hypos.mgz',  {'raw_image':'aseg.presurf.hypos', 'load': _load_imm_mgh},
        'aseg.presurf.mgz',        ({'raw_image':'aseg.presurf', 'load': _load_imm_mgh},
                                    {'image':'presurface_segmentation', 'load': _load_imm_mgh}), 
        'brain.mgz',               ({'raw_image':'brain', 'load': _load_imm_mgh},
                                    {'image':'brain', 'load': _load_imm_mgh}),
        'brain.finalsurfs.mgz',    {'raw_image':'brain.finalsurfs', 'load': _load_imm_mgh},
        'wm.seg.mgz',              {'raw_image':'wm.seg', 'load': _load_imm_mgh},
        'wm.asegedit.mgz',         {'raw_image':'wm.asegedit', 'load': _load_imm_mgh},
        'wm.mgz',                  ({'raw_image':'wm', 'load': _load_imm_mgh},
                                    {'image':'white_matter', 'load': _load_imm_mgh}),
        'filled.mgz',              {'raw_image':'filled', 'load': _load_imm_mgh},
        'T2.mgz',                  ({'raw_image':'T2', 'load': _load_imm_mgh},
                                    {'image':'T2', 'load': _load_imm_mgh}),
        'ribbon.mgz',              ({'raw_image':'ribbon', 'load': _load_imm_mgh},
                                    {'image':'ribbon', 'load': _load_imm_mgh}),
        'aparc+aseg.mgz',          ({'raw_image':'aparc+aseg', 'load': _load_imm_mgh},
                                    {'image':'Desikan06_parcellation', 'load': _load_imm_mgh}),
        'aparc.DKTatlas+aseg.mgz', ({'raw_image':'aparc.DKTatlas+aseg', 'load': _load_imm_mgh},
                                    {'image':'DKT40_parcellation', 'load': _load_imm_mgh}),
        'aparc.a2009s+aseg.mgz',   ({'raw_image':'aparc.a2009s+aseg', 'load': _load_imm_mgh},
                                    {'image':'Destrieux09_parcellation', 'load': _load_imm_mgh},
                                    {'image':'parcellation', 'load': _load_imm_mgh}),
        'ctrl_pts.mgz',            {'raw_image':'ctrl_pts', 'load': _load_imm_mgh},
        'lh.ribbon.mgz',           ({'raw_image':'lh.ribbon', 'load': _load_imm_mgh},
                                    {'image':'lh_ribbon', 'load': _load_imm_mgh}),
        'rh.ribbon.mgz',           ({'raw_image':'rh.ribbon', 'load': _load_imm_mgh},
                                    {'image':'rh_ribbon', 'load': _load_imm_mgh}),
        'wmparc.mgz',              ({'raw_image':'wmparc', 'load': _load_imm_mgh},
                                    {'image':'white_parcellation', 'load': _load_imm_mgh}),
        'aseg.mgz',                ({'raw_image':'aseg', 'load': _load_imm_mgh},
                                    {'image':'segmentation', 'load': _load_imm_mgh})],
    'surf', [
        'lh.area',           ({'hemi':'lh',  'property':'white_surface_area', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'white_surface_area', 'load':_load_prop}),
        'lh.area.mid',       ({'hemi':'lh',  'property':'midgray_surface_area', 'load':_load_prop},
                              {'hemi':'lh',  'property':'surface_area', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'midgray_surface_area', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'surface_area', 'load':_load_prop}),
        'lh.area.pial',      ({'hemi':'lh',  'property':'pial_surface_area', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'pial_surface_area', 'load':_load_prop}),
        'lh.avg_curv',       ({'hemi':'lh',  'property':'atlas_curvature', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'atlas_curvature', 'load':_load_prop}),
        'lh.curv',           ({'hemi':'lh',  'property':'white_curvature', 'load':_load_prop},
                              {'hemi':'lh',  'property':'curvature', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'white_curvature', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'curvature', 'load':_load_prop}),
        'lh.curv.pial',      ({'hemi':'lh',  'property':'pial_curvature', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'pial_curvature', 'load':_load_prop}),
        'lh.jacobian_white', ({'hemi':'lh',  'property':'jacobian_norm', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'jacobian_norm', 'load':_load_prop}),
        'lh.inflated',       {'hemi':'lh',  'surface':'inflated', 'load':_load_surfx},
        'lh.orig',           {'hemi':'lh',  'surface':'white_original', 'load':_load_surfx},
        'lh.pial',           {'hemi':'lh',  'surface':'pial', 'load':_load_surfx},
        'lh.smoothwm',       {'hemi':'lh',  'surface':'white_smooth', 'load':_load_surfx},
        'lh.sphere',         ({'hemi':'lh',  'surface':'sphere', 'load':_load_surfx},
                              {'hemi':'lh',  'registration':'native', 'load':_load_surfx}),
        'lh.sphere.reg',     {'hemi':'lh',  'registration':'fsaverage', 'load':_load_surfx},
        'lh.white',          ({'hemi':'lh',  'surface':'white',
                               'load':_load_geo, 'filt':lambda u:u[0]},
                              {'hemi':'lh',  'tess':'white',
                               'load':_load_geo, 'filt':lambda u:u[1]}),
        'lh.sulc',           ({'hemi':'lh',  'property':'convexity', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'convexity', 'load':_load_prop}),
        'lh.thickness',      ({'hemi':'lh',  'property':'thickness', 'load':_load_prop},
                              {'hemi':'rhx', 'property':'thickness', 'load':_load_prop}),
        'rh.area',           ({'hemi':'rh',  'property':'white_surface_area', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'white_surface_area', 'load':_load_prop}),
        'rh.area.mid',       ({'hemi':'rh',  'property':'midgray_surface_area', 'load':_load_prop},
                              {'hemi':'rh',  'property':'surface_area', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'midgray_surface_area', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'surface_area', 'load':_load_prop}),
        'rh.area.pial',      ({'hemi':'rh',  'property':'pial_surface_area', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'pial_surface_area', 'load':_load_prop}),
        'rh.avg_curv',       ({'hemi':'rh',  'property':'atlas_curvature', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'atlas_curvature', 'load':_load_prop}),
        'rh.curv',           ({'hemi':'rh',  'property':'white_curvature', 'load':_load_prop},
                              {'hemi':'rh',  'property':'curvature', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'white_curvature', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'curvature', 'load':_load_prop}),
        'rh.curv.pial',      ({'hemi':'rh',  'property':'pial_curvature', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'pial_curvature', 'load':_load_prop}),
        'rh.inflated',       ({'hemi':'rh',  'surface':'inflated', 'load':_load_surfx},
                              {'hemi':'lhx', 'surface':'inflated', 'load':_load_surfx}),
        'rh.jacobian_white', ({'hemi':'rh',  'property':'jacobian_norm', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'jacobian_norm', 'load':_load_prop}),
        'rh.orig',           ({'hemi':'rh',  'surface':'white_original', 'load':_load_surfx},
                              {'hemi':'lhx', 'surface':'white_original', 'load':_load_surfx}),
        'rh.pial',           ({'hemi':'rh',  'surface':'pial', 'load':_load_surfx},
                              {'hemi':'lhx', 'surface':'pial', 'load':_load_surfx}),
        'rh.smoothwm',       ({'hemi':'rh',  'surface':'white_smooth', 'load':_load_surfx},
                              {'hemi':'lhx', 'surface':'white_smooth', 'load':_load_surfx}),
        'rh.sphere',         ({'hemi':'rh',  'surface':'sphere', 'load':_load_surfx},
                              {'hemi':'rh',  'registration':'native', 'load':_load_surfx},
                              {'hemi':'lhx', 'surface':'sphere', 'load':_load_surfx},
                              {'hemi':'lhx', 'registration':'native', 'load':_load_surfx}),
        'rh.sphere.reg',     ({'hemi':'rh',  'registration':'fsaverage', 'load':_load_surfx},
                              {'hemi':'lhx', 'registration':'fsaverage', 'load':_load_surfx}),
        'rh.sulc',           ({'hemi':'rh',  'property':'convexity', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'convexity', 'load':_load_prop}),
        'rh.thickness',      ({'hemi':'rh',  'property':'thickness', 'load':_load_prop},
                              {'hemi':'lhx', 'property':'thickness', 'load':_load_prop}),
        'rh.white',          ({'hemi':'rh',  'surface':'white',
                               'load':_load_geo, 'filt':lambda u:u[0]},
                              {'hemi':'rh',  'tess':'white',
                               'load':_load_geo, 'filt':lambda u:u[1]},
                              {'hemi':'lhx', 'surface':'white',
                               'load':_load_geo, 'filt':lambda u:u[0]},
                              {'hemi':'lhx', 'tess':'white',
                               'load':_load_geo, 'filt':lambda u:u[1]}),
        # a few other registrations that we recognize
        'lh.fsaverage_sym.sphere.reg',        {'hemi':'lh',  'registration':'fsaverage_sym',
                                               'load':_load_surfx},
        #'rh.fsaverage_sym.sphere.reg',        {'hemi':'lhx', 'registration':'fsaverage_sym',
        #                                       'load':_load_surfx},
        'lh.benson14_retinotopy.sphere.reg',  ({'hemi':'lh', 'registration':'benson14_retinotopy',
                                                'load':_load_surfx},
                                               {'hemi':'lh', 'registration':'benson14',
                                                'load':_load_surfx}),
        'rh.benson14_retinotopy.sphere.reg',  ({'hemi':'rh', 'registration':'benson14_retinotopy',
                                                'load':_load_surfx},
                                               {'hemi':'rh', 'registration':'benson14',
                                                'load':_load_surfx}),
        'lh.benson17_retinotopy.sphere.reg',  ({'hemi':'lh', 'registration':'benson17_retinotopy',
                                                'load':_load_surfx},
                                               {'hemi':'lh', 'registration':'benson17',
                                                'load':_load_surfx}),
        'rh.benson17_retinotopy.sphere.reg',  ({'hemi':'rh', 'registration':'benson17_retinotopy',
                                                'load':_load_surfx},
                                               {'hemi':'rh', 'registration':'benson17',
                                                'load':_load_surfx})],
    'xhemi', [
        'surf', [
            #'rh.fsaverage_sym.sphere.reg',        {'hemi':'lhx', 'registration':'fsaverage_sym',
            #                                   'load':_load_surfx},
            'lh.fsaverage_sym.sphere.reg',    {'hemi':'rh',  'registration':'fsaverage_sym',
                                               'load':_load_surfx},
            'lh.inflated',   {'hemi':'rhx',  'surface':'inflated', 'load':_load_surfx},
            'lh.orig',       {'hemi':'rhx',  'surface':'white_original', 'load':_load_surfx},
            'lh.pial',       {'hemi':'rhx',  'surface':'pial', 'load':_load_surfx},
            'lh.smoothwm',   {'hemi':'rhx',  'surface':'white_smooth', 'load':_load_surfx},
            'lh.sphere',     ({'hemi':'rhx',  'surface':'sphere', 'load':_load_surfx},
                              {'hemi':'rhx',  'registration':'native', 'load':_load_surfx}),
            'lh.sphere.reg', {'hemi':'rhx',  'registration':'fsaverage', 'load':_load_surfx},
            'lh.white',      ({'hemi':'rhx',  'surface':'white',
                               'load':_load_geo, 'filt':lambda u:u[0]},
                              {'hemi':'rhx',  'tess':'white',
                               'load':_load_geo, 'filt':lambda u:u[1]}),
            'rh.inflated',   {'hemi':'lhx',  'surface':'inflated', 'load':_load_surfx},
            'rh.orig',       {'hemi':'lhx',  'surface':'white_original', 'load':_load_surfx},
            'rh.pial',       {'hemi':'lhx',  'surface':'pial', 'load':_load_surfx},
            'rh.smoothwm',   {'hemi':'lhx',  'surface':'white_smooth', 'load':_load_surfx},
            'rh.sphere',     ({'hemi':'lhx',  'surface':'sphere', 'load':_load_surfx},
                             {'hemi':'lhx',  'registration':'native', 'load':_load_surfx}),
            'rh.sphere.reg', {'hemi':'lhx',  'registration':'fsaverage', 'load':_load_surfx},
            'rh.white',      ({'hemi':'lhx',  'surface':'white',
                               'load':_load_geo, 'filt':lambda u:u[0]},
                              {'hemi':'lhx',  'tess':'white',
                               'load':_load_geo, 'filt':lambda u:u[1]})]],
    'label', [
        # cortex labels:
        'lh.cortex.label',           ({'hemi':'lh',  'label':'cortex', 'load':_load_label},
                                      {'hemi':'rhx', 'label':'cortex', 'load':_load_label}),
        'rh.cortex.label',           ({'hemi':'rh',  'label':'cortex', 'load':_load_label},
                                      {'hemi':'lhx', 'label':'cortex', 'load':_load_label}),
        # the annotation files:
        'lh.BA_exvivo.annot',        ({'hemi':'lh',  'annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'lh.BA_exvivo.thresh.annot', ({'hemi':'lh',  'annot':'brodmann_area', 'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'brodmann_area', 'load':_load_annot}),
        'lh.BA.annot',               ({'hemi':'lh',  'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'lh.BA.thresh.annot',        ({'hemi':'lh',  'alt_annot':'brodmann_area',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'alt_annot':'brodmann_area',
                                       'load':_load_annot}),
        'lh.aparc.annot',            ({'hemi':'lh',  'annot':'Desikan06_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'Desikan06_parcellation',
                                       'load':_load_annot}),
        'lh.aparc.a2009s.annot',     ({'hemi':'lh',  'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lh',  'annot':'parcellation', 'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'parcellation', 'load':_load_annot}),
        'lh.aparc.DKTatlas.annot',   ({'hemi':'lh',  'annot':'DKT40_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'DKT40_parcellation',
                                       'load':_load_annot}),
        'lh.BA_exvivo.annot',        ({'hemi':'lh',  'annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'lh.BA_exvivo.thresh.annot', ({'hemi':'lh',  'annot':'brodmann_area', 'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'brodmann_area', 'load':_load_annot}),
        'lh.BA.annot',               ({'hemi':'lh',  'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'lh.BA.thresh.annot',        ({'hemi':'lh',  'alt_annot':'brodmann_area',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'alt_annot':'brodmann_area',
                                       'load':_load_annot}),
        'lh.aparc.annot',            ({'hemi':'lh',  'annot':'Desikan06_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'Desikan06_parcellation',
                                       'load':_load_annot}),
        'lh.aparc.a2009s.annot',     ({'hemi':'lh', 'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lh', 'annot':'parcellation', 'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'parcellation', 'load':_load_annot}),
        'lh.aparc.DKTatlas.annot',   ({'hemi':'lh',  'annot':'DKT40_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rhx', 'annot':'DKT40_parcellation',
                                       'load':_load_annot}),
        'rh.BA_exvivo.annot',        ({'hemi':'rh',  'annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'rh.BA_exvivo.thresh.annot', ({'hemi':'rh',  'annot':'brodmann_area', 'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'brodmann_area', 'load':_load_annot}),
        'rh.BA.annot',               ({'hemi':'rh',  'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'rh.BA.thresh.annot',        ({'hemi':'rh',  'alt_annot':'brodmann_area',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'alt_annot':'brodmann_area',
                                       'load':_load_annot}),
        'rh.aparc.annot',            ({'hemi':'rh',  'annot':'Desikan06_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'Desikan06_parcellation',
                                       'load':_load_annot}),
        'rh.aparc.a2009s.annot',     ({'hemi':'rh',  'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rh',  'annot':'parcellation', 'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'parcellation', 'load':_load_annot}),
        'rh.aparc.DKTatlas.annot',   ({'hemi':'rh',  'annot':'DKT40_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'DKT40_parcellation',
                                       'load':_load_annot}),
        'rh.BA_exvivo.annot',        ({'hemi':'rh',  'annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'rh.BA_exvivo.thresh.annot', ({'hemi':'rh',  'annot':'brodmann_area', 'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'brodmann_area', 'load':_load_annot}),
        'rh.BA.annot',               ({'hemi':'rh',  'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'alt_annot':'brodmann_area_wide',
                                       'load':_load_annot}),
        'rh.BA.thresh.annot',        ({'hemi':'rh',  'alt_annot':'brodmann_area',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'alt_annot':'brodmann_area',
                                       'load':_load_annot}),
        'rh.aparc.annot',            ({'hemi':'rh',  'annot':'Desikan06_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'Desikan06_parcellation',
                                       'load':_load_annot}),
        'rh.aparc.a2009s.annot',     ({'hemi':'rh', 'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'rh', 'annot':'parcellation', 'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'Destrieux09_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'parcellation', 'load':_load_annot}),
        'rh.aparc.DKTatlas.annot',   ({'hemi':'rh',  'annot':'DKT40_parcellation',
                                       'load':_load_annot},
                                      {'hemi':'lhx', 'annot':'DKT40_parcellation',
                                       'load':_load_annot})
    ] + [
        x
        for (h,hx) in zip(['lh','rh'],['rhx','lhx']) for p in (True, False) for alt in (True,False)
        for s in ['BA1', 'BA2', 'BA3a', 'BA3b', 'BA44', 'BA45', 'BA4a', 'BA4p', 'BA6',
                  'MT', 'V1', 'V2', 'entorhinal', 'perirhinal']
        for x in ['%s.%s%s.%slabel' % (h, s, '' if alt else '_exvivo', '' if p else 'thresh.'),
                  ({'hemi':h, (('alt_' if alt else '') + ('weight' if p else 'label')):s,
                    'load':_load_label},
                  {'hemi':hx, (('alt_' if alt else '') + ('weight' if p else 'label')):s,
                    'load':_load_label})]]]
def subject_file_map(path):
    '''
    subject_file_map(path) yields a filemap object for the given freesurfer subject path.
    '''
    return file_map(path, freesurfer_subject_filemap_instructions,
                    data_hierarchy=freesurfer_subject_data_hierarchy)
_registration_aliases = {'benson14_retinotopy.v4_0': ['benson17', 'benson17_retinotopy']}
def cortex_from_filemap(fmap, chirality, name, subid=None, affine=None):
    '''
    cortex_from_filemap(filemap, chirality, name) yields a cortex object from the given filemap;
      if the chirality and name do not match, then the result must be an xhemi.
    '''
    chirality = chirality.lower()
    name = name.lower()
    # get the relevant hemi-data
    hdat = fmap.data_tree.hemi[name]
    # we need the tesselation at build-time, so let's create that now:
    tris = hdat.tess['white']
    # this tells us the max number of vertices
    n = np.max(tris) + 1
    # Properties: we want to merge a bunch of things together...
    # for labels, weights, annots, we need to worry about loading alts:
    def _load_with_alt(k, s0, sa, trfn):
        try: u = s0.get(k, None)
        except Exception: u = None
        if u is None and sa is not None:
            try: u = sa.get(k, None)
            except Exception: u = None
        if u is None: raise ValueError('Exception while loading property %s' % k)
        else: return u if trfn is None else trfn(u)
    def _lbltr(ll):
        l = np.zeros(n, dtype='bool')
        l[ll[0]] = True
        l.setflags(write=False)
        return l
    def _wgttr(ll):
        w = np.zeros(n, dtype='float')
        w[ll[0]] = ll[1]
        w.setflags(write=False)
        return w
    def _anotr(ll):
        ll[0].setflags(write=False)
        return ll[0]
    p = {}
    from itertools import chain
    l  = hdat.label     if hasattr(hdat, 'label')     else {}
    al = hdat.alt_label if hasattr(hdat, 'alt_label') else {}
    for k in set(chain(six.iterkeys(l), six.iterkeys(al))):
        p[k+'_label'] = curry(_load_with_alt, k, l, al, _lbltr)
    w  = hdat.weight     if hasattr(hdat, 'weight')     else {}
    aw = hdat.alt_weight if hasattr(hdat, 'alt_weight') else {}
    for k in set(chain(six.iterkeys(w), six.iterkeys(aw))):
        p[k+'_weight'] = curry(_load_with_alt, k, w, aw, _wgttr)
    a  = hdat.annot     if hasattr(hdat, 'annot')     else {}
    aa = hdat.alt_annot if hasattr(hdat, 'alt_annot') else {}
    for k in set(chain(six.iterkeys(a), six.iterkeys(aa))):
        p[k] = curry(_load_with_alt, k, a, aa, _anotr)
    props = pimms.merge(hdat.property, pimms.lazy_map(p))
    tess = geo.Tesselation(tris, properties=props)
    def _make_midgray():
        x = np.mean([hdat.surface['white'], hdat.surface['pial']], axis=0)
        x.setflags(write=False)
        return x
    surfs = hdat.surface.set('midgray', _make_midgray)
    # if this is a subject that exists in the library, we may want to add some files:
    if subid is None:
        pd = fmap.pseudo_paths[None]._path_data
        subid = pd['pathmod'].split(fmap.actual_path)[1]
    regs = hdat.registration
    extra_path = os.path.join(library_freesurfer_subjects(), subid, 'surf')
    if os.path.isdir(extra_path):
        for flnm in os.listdir(extra_path):
            if flnm.startswith(chirality + '.') and flnm.endswith('.sphere.reg'):
                mid = flnm[(len(chirality)+1):-11]
                if mid == '': mid = ['fsaverage']
                else:         mid = _registration_aliases.get(mid, [mid])
                load_mid_fn = curry(_load_surfx, os.path.join(extra_path, flnm), {})
                for mid in mid:
                    f = load_mid_fn if mid not in regs else \
                        curry(try_until,
                              curry(lambda regs: regs[mid], regs), load_mid_fn,
                              check=pimms.is_matrix)
                    regs = regs.set(mid, f)

    # Okay, make the cortex object!
    md = {'file_map': fmap}
    if subid is not None: md['subject_id'] = subid
    return mri.Cortex(chirality, tess, surfs, regs, affine=affine, meta_data=md).persist()
def images_from_filemap(fmap):
    '''
    images_from_filemap(fmap) yields a persistent map of MRImages tracked by the given subject with
      the given name and path; in freesurfer subjects these are renamed and converted from their
      typical freesurfer filenames (such as 'ribbon') to forms that conform to the neuropythy naming
      conventions (such as 'gray_mask'). To access data by their original names, use the filemap.
    '''
    ims = {}
    raw_images = fmap.data_tree.raw_image
    def _make_imm_mask(rib, val, eq=True):
        arr = np.asarray(rib.dataobj)
        arr = (arr == val) if eq else (arr != val)
        arr.setflags(write=False)
        return fsmgh.MGHImage(arr, rib.affine, rib.header)
    # start with the ribbon:
    ims['lh_gray_mask']  = lambda:_make_imm_mask(raw_images['ribbon'], 3)
    ims['lh_white_mask'] = lambda:_make_imm_mask(raw_images['ribbon'], 2)
    ims['rh_gray_mask']  = lambda:_make_imm_mask(raw_images['ribbon'], 42)
    ims['rh_white_mask'] = lambda:_make_imm_mask(raw_images['ribbon'], 41)
    ims['brain_mask']    = lambda:_make_imm_mask(raw_images['ribbon'], 0, False)
    # merge in with the typical images
    return pimms.merge(fmap.data_tree.image, pimms.lazy_map(ims))
def subject_from_filemap(fmap, name=None, meta_data=None, check_path=True):
    # start by making a pseudo-dir:
    if check_path and not is_freesurfer_subject_path(fmap.pseudo_paths[None]):
        raise ValueError('given path does not appear to hold a freesurfer subject')
    # we need to go ahead and load the ribbon...
    rawims = fmap.data_tree.raw_image
    rib = next(rawims[k]
               for k in ['ribbon', 'brain', 'intensity_normalized', 'conformed']
               if k in rawims and rawims[k] is not None)
    vox2nat = rib.affine
    vox2vtx = rib.header.get_vox2ras_tkr()
    vtx2nat = np.dot(vox2nat, np.linalg.inv(vox2vtx))
    # make images and hems
    imgs = images_from_filemap(fmap)
    hems = pimms.lazy_map({h:curry(cortex_from_filemap, fmap, h, ch, name, vtx2nat)
                           for (h,ch) in zip(['lh','rhx','rh','lhx'], ['lh','lh','rh','rh'])})
    meta_data = pimms.persist({} if meta_data is None else meta_data)
    meta_data = meta_data.set('raw_images', fmap.data_tree.raw_image)
    return mri.Subject(name=name, pseudo_path=fmap.pseudo_paths[None],
                       hemis=hems, images=imgs,
                       meta_data=meta_data)
@nyio.importer('freesurfer_subject', sniff=is_freesurfer_subject_path)
def subject(path, name=Ellipsis, meta_data=None, check_path=True, filter=None):
    '''
    subject(name) yields a freesurfer Subject object for the subject with the given name. Subjects
      are cached and not reloaded, so multiple calls to subject(name) will yield the same immutable
      subject object.

    Note that subects returned by freesurfer_subject() are always persistent Immutable objects; this
    means that you must create a transient version of the subject to modify it via the member
    function sub.transient(). Better, you can make copies of the objects with desired modifications
    using the copy method--see the pimms library documentation regarding immutable classes and
    objects.

    If you wish to modify all subjects loaded by this function, you may set its attribute 'filter'
    to a function or list of functions that take a single argument (a subject object) and returns a
    single argument (a potentially-modified subject object).

    The argument name may alternately be a pseudo_path object or a path that can be converted into a
    pseudo_path object.

    The following options are accepted:
      * name (default: Ellipsis) may optionally specify the subject's name; if Ellipsis, then
        attempts to deduce the name from the initial argument (which may be a name or a path).
      * meta_data (default: None) may optionally be a map that contains meta-data to be passed along
        to the subject object (note that this meta-data will not be cached).
      * check_path (default: True) may optionally be set to False to ignore the requirement that a
        directory contain at least the mri/, label/, and surf/ directories to be considered a valid
        FreeSurfer subject directory. Subject objects returned using this argument are not cached.
        Additionally, check_path may be set to None instead of False, indicating that no checks or
        search should be performed; the string name should be trusted to be an exact relative or
        absolute path to a valid FreeSurfer subejct.
      * filter (default: None) may optionally specify a filter that should be applied to the subject
        before returning. This must be a function that accepts as an argument the subject object and
        returns a (potentially) modified subject object. Filtered subjects are cached by the id
        of the filters.
    '''
    # convert the path to a pseudo-dir; this may fail if the user is requesting a subject by name...
    try: pdir = to_pseudo_path(path)
    except Exception: pdir = None
    if pdir is None: # search for a subject with this name
        tmp = find_subject_path(path, check_path=check_path)
        if tmp is not None:
            pdir = to_pseudo_path(tmp)
            path = tmp
        elif path == 'fsaverage':
            try:
                import neuropythy as ny
                return ny.data['benson_winawer_2018'].subjects['fsaverage']
            except: pass
    if pdir is None: raise ValueError('could not find freesurfer subject: %s' % path)
    path = pdir.source_path
    # okay, before we continue, lets check the cache...
    if path in subject._cache: sub = subject._cache[path]
    else:
        # make the filemap
        fmap = subject_file_map(pdir)
        # extract the name if need-be
        if name is Ellipsis:
            import re
            (pth,name) = (path, '.')
            while name == '.': (pth, name) = pdir._path_data['pathmod'].split(pth)
            name = name.split(':')[-1]
            name = pdir._path_data['pathmod'].split(name)[1]
            if '.tar' in name: name = name.split('.tar')[0]
        # and make the subject!
        sub = subject_from_filemap(fmap, name=name, check_path=check_path, meta_data=meta_data)
        if mri.is_subject(sub):
            sub.persist()
            sub = sub.with_meta(file_map=fmap)
            subject._cache[path] = sub
    # okay, we have the initial subject; let's organize the filters
    if pimms.is_list(subject.filter) or pimms.is_tuple(subject.filter): filts = list(subject.filter)
    else: filts = []
    if pimms.is_list(filter) or pimms.is_tuple(filter): filter = list(filter)
    else: filter = []
    filts = filts + filter
    if len(filts) == 0: return sub
    fids = tuple([id(f) for f in filts])
    tup = fids + (path,)
    if tup in subject._cache: return subject._cache[tup]
    for f in filts: sub = f(sub)
    if mri.is_subject(sub): subject._cache[tup] = sub
    return sub
subject._cache = {}
subject.filter = None
def forget_subject(sid):
    '''
    forget_subject(sid) causes neuropythy's freesurfer module to forget about cached data for the
      subject with subject id sid. The sid may be any sid that can be passed to the subject()
      function.

    This function is useful for batch-processing of subjects in a memory-limited environment; e.g.,
    if you run out of memory while processing FreeSurfer subjects it is possibly because neuropythy
    is caching all of their data instead of freeing it.
    '''
    sub = subject(sid)
    if sub.path in subject._cache:
        del subject._cache[sub.path]
    for (k,v) in six.iteritems(subject._cache):
        if pimms.is_tuple(k) and k[-1] == sub.path:
            del subject._cache[k]
    return None
def forget_all():
    '''
    forget_all() causes neuropythy's freesurfer module to forget all cached subjects. See also
    forget_subject.
    '''
    subject._cache = {}
    return None

# This function creates the tkr matrix for a volume given the dims
def tkr_vox2ras(img, zooms=None):
    '''
    tkr_vox2ras(img) yields the FreeSurfer tkr VOX2RAS matrix for the given nibabel image object
      img. The img must have a get_shape() method and header member with a get_zooms() method.
    tkr_vox2ras(hdr) operates on a nibabel image header object.
    tkr_vox2ras(shape, zooms) operates on the shape (e.g., for FreeSurfer subjects (256,256,256))
      and the zooms or voxel dimensions (e.g., for FreeSurfer subjects, (1.0, 1.0, 1.0)).
    '''
    if zooms is not None:
        # let's assume that they passed shape, zooms
        shape = img
    else:
        try:              img = img.header
        except Exception: pass
        try:              (shape, zooms) = (img.get_data_shape(), img.get_zooms())
        except Exception: raise ValueError('single argument must be nibabel image or header')
    # Okay, we have shape and zooms...
    zooms = zooms[0:3]
    shape = shape[0:3]
    (dC, dR, dS) = zooms
    (nC, nR, nS) = 0.5 * (np.asarray(shape) * zooms)
    return np.asarray([[-dC,   0,   0,  nC],
                       [  0,   0,  dS, -nS],
                       [  0, -dR,   0,  nR],
                       [  0,   0,   0,   1]])

class SubjectDir(collections.Mapping):
    '''
    SubjectsDir objects are dictionary-like objects that represent a particular subject directory.
    They satisfy their queries (such as `'bert' in spath`) by querying the filesystem itself.

    For more information see the subjects_path function.
    '''
    def __init__(self, path, bids=False, filter=None, meta_data=None, check_path=True):
        path = os.path.expanduser(os.path.expandvars(path))
        self.path = os.path.abspath(path)
        if not os.path.isdir(self.path): raise ValueError('given path is not a directory')
        self.bids = bool(bids)
        self.options = dict(filter=filter, meta_data=meta_data, check_path=bool(check_path))
    def __contains__(self, sub):
        # first check the item straight-up:
        sd = os.path.join(self.path, sub)
        if os.path.isdir(sd) and (not check_path or is_freesurfer_subject_path(sd)): return True
        if not bids: return False
        if sub.startswith('sub-'): sd = os.path.join(self.path, sub[4:])
        else: sd = os.path.join(self.path, 'sub-' + sub)
        return os.path.isdir(sd) and (not check_path or is_freesurfer_subject_path(sd))
    def _get_subject(self, sd, name):
        opts = dict(**self.options)
        opts['name'] = name
        return subject(sd, **opts)
    def __getitem__(self, sub):
        check_path = self.options['check_path']
        if self.bids:
            if sub.startswith('sub-'): (sub, name) = (sub, name[4:])
            else: (sub,name) = ('sub-' + sub, sub)
        else: name = sub
        sd = os.path.join(self.path, sub)
        if os.path.isdir(sd) and (not check_path or is_freesurfer_subject_path(sd)):
            return self._get_subject(sd, name)
        if not bids: return False
        # try without the 'sub-' (e.g. for fsaverage)
        sub = name
        sd = os.path.join(self.path, sub)
        if os.path.isdir(sd) and (not check_path or is_freesurfer_subject_path(sd)):
            return self._get_subject(sd, name)
        raise KeyError(sub)
    def asmap(self):
        check_path = self.options['check_path']
        # search through the subjects in this directory
        res = {}
        for sub in os.listdir(self.path):
            if self.bids and sub.startswith('sub-'): (sub,name) = (sub,sub[4:])
            else: name = sub
            sdir = os.path.join(self.path, sub)
            if not check_path or is_freesurfer_subject_path(sdir):
                res[name] = curry(self._get_subject, sdir, name)
        return pimms.lmap(res)
    def __len__(self): return len(self.asmap())
    def __iter__(self): return iter(self.asmap())
    def __repr__(self): return 'freesurfer.SubjectDir(' + repr(self.asmap()) + ')'
    def __str__(self): return 'freesurfer.SubjectDir(' + str(self.asmap()) + ')'
# Functions for handling freesurfer subject directories themselves
def is_freesurfer_subject_dir_path(path):
    '''
    is_freesurfer_subject_dir_path(path) yields True if path is a directory that contains at least
      one FreeSurfer subejct subdirectory.
    '''
    if not os.path.isdir(path): return False
    for p in os.listdir(path):
        pp = os.path.join(path, p)
        if not os.path.isdir(pp): continue
        if is_freesurfer_subject_path(pp): return True
    return False
@nyio.importer('freesurfer_subject_dir', sniff=is_freesurfer_subject_dir_path)
def subject_dir(path, bids=False, filter=None, meta_data=None, check_path=True):
    '''
    subject_dir(path yields a dictionary-like object containing the subjects in the FreeSurfer
      subjects directory given by path.

    The following optional arguments are accepted:
      * bids (default: False) may be set to True to indicate that the directory is part of a BIDS
        directory; in this case the 'sub-' prefix is stripped from directory names to deduce subject
        names.
      * check_path (default: True) may optionally be set to False to ignore the requirement that a
        subject directory contain at least the mri/, label/, and surf/ directories to be considered
        a valid FreeSurfer subject directory. See help(neuropythy.freesurfer.subject) for more
        information.
      * filter (default: None) may optionally specify a filter that should be applied to the subject
        before returning. See help(neuropythy.freesurfer.subject) for more information.
      * meta_data (default: None) may optionally be a map that contains meta-data to be passed along
        to the subject object (note that this meta-data will not be cached).
    '''
    return SubjectDir(path, bids=bids, filter=filter, meta_data=meta_data, check_path=check_path)

####################################################################################################
# import/export code that works with neuropythy.io

# MGH Images!
@nyio.importer('mgh', ('mgh', 'mgh.gz', 'mgz'))
def load_mgh(filename, to='auto'):
    '''
    load_mgh(filename) yields the MGHImage referened by the given filename by using the
      nibabel.freesurfer.mghformat.load function.
    
    The optional argument 'to' may be used to coerce the resulting data to a particular format; the
    following arguments are understood:
      * 'header' will yield just the image header
      * 'data' will yield the image's data-array
      * 'field' will yield a squeezed version of the image's data-array and will raise an error if
        the data object has more than 2 non-unitary dimensions (appropriate for loading surface
        properties stored in image files)
      * 'affine' will yield the image's affine transformation
      * 'image' will yield the raw image object
      * 'auto' is equivalent to 'image' unless the image has no more than 2 non-unitary dimensions,
        in which case it is assumed to be a surface-field and the return value is equivalent to
        the 'field' value.
    '''
    img = fsmgh.load(filename)
    to = to.lower()
    if to == 'image':    return img
    elif to == 'data':   return img.dataobj
    elif to == 'affine': return img.affine
    elif to == 'header': return img.header
    elif to == 'field':
        dat = np.squeeze(img.dataobj)
        if len(dat.shape) > 2:
            raise ValueError('image requested as field has more than 2 non-unitary dimensions')
        return dat
    elif to in ['auto', 'automatic']:
        dims = set(img.dataobj.shape)
        if 1 < len(dims) < 4 and 1 in dims:
            return np.squeeze(img.dataobj)
        else:
            return img
    else:
        raise ValueError('unrecognized \'to\' argument \'%s\'' % to)
def to_mgh(obj, like=None, header=None, affine=None, extra=Ellipsis):
    '''
    to_mgh(obj) yields an MGHmage object that is as equivalent as possible to the given object obj.
      If obj is an MGHImage already and no changes are requested, then it is returned unmolested;
      otherwise, the optional arguments can be used to edit the header, affine, and exta.

    The following options are accepted:
      * like (default: None) may be provided to give a guide for the various header- and meta-data
        that is included in the image. If this is a nifti image object, its meta-data are used; if
        this is a subject, then the meta-data are deduced from the subject's voxel and native
        orientation matrices. All other specific options below override anything deduced from the
        like argument.
      * header (default: None) may be a Nifti1 or Niti2 image header to be used as the nifti header
        or to replace the header in a new image.
      * affine (default: None) may specify the affine transform to be given to the image object.
      * extra (default: Ellipsis) may specify an extra mapping for storage with the image data; the
        default value Ellipsis indicates that such a mapping should be left as is if it already
        exists in the obj or in the like option and should otherwise be empty.
    '''
    obj0 = obj
    # First go from like to explicit versions of affine and header:
    if like is not None:
        if isinstance(like, nib.analyze.AnalyzeHeader) or isinstance(obj, fsmgh.MGHHeader):
            if header is None: header = like
        elif isinstance(like, nib.analyze.SpatialImage):
            if header is None: header = like.header
            if affine is None: affine = like.affine
            if extra is Ellipsis: extra = like.extra
        elif isinstance(like, mri.Subject):
            if affine is None: affine = like.images['brain'].affine
        else:
            raise ValueError('Could not interpret like argument with type %s' % type(like))
    # check to make sure that we have to change something:
    elif (isinstance(obj, fsmgh.MGHImage)):
        if ((header is None or obj.header is header) and
            (extra is Ellipsis or extra == obj.extra or (extra is None and len(obj.extra) == 0))):
            return obj
    # okay, now look at the header and affine etc.
    if header is None:
        if isinstance(obj, nib.analyze.SpatialImage):
            header = obj.header
        else:
            header = None
    if affine is None:
        if isinstance(obj, nib.analyze.SpatialImage):
            affine = obj.affine
        else:
            affine = np.eye(4)
    if extra is None: extra = {}
    # Figure out what the data is
    if isinstance(obj, nib.analyze.SpatialImage):
        obj = obj.dataobj
    else:
        obj = np.asarray(obj)
    if len(obj.shape) < 3: obj = np.asarray([[obj]])
    # make sure the dtype isn't a high-bit integer or float
    if np.issubdtype(obj.dtype, np.integer) and obj.dtype.itemsize > 4:
        obj = np.asarray(obj, dtype=np.int32)
    elif np.issubdtype(obj.dtype, np.floating) and obj.dtype.itemsize > 4:
        obj = np.asarray(obj, dtype=np.float32)
    # Okay, make a new object now...
    obj = fsmgh.MGHImage(obj, affine, header=header, extra=extra)
    # Okay, that's it!
    return obj
@nyio.exporter('mgh', ('mgh', 'mgh.gz', 'mgz'))
def save_mgh(filename, obj, like=None, header=None, affine=None, extra=Ellipsis):
    '''
    save_mgh(filename, obj) saves the given object to the given filename in the mgh format and
      returns the filename.

    All options that can be given to the to_mgh function can also be passed to this function; they
    are used to modify the object prior to exporting it.
    '''
    obj = to_mgh(obj, like=like, header=header, affine=affine, extra=extra)
    obj.to_filename(filename)
    return filename

# For other freesurfer formats
@nyio.importer('freesurfer_geometry', ('white', 'pial', 'sphere', 'sphere.reg', 'inflated'))
def load_freesurfer_geometry(filename, to='mesh', warn=False):
    '''
    load_freesurfer_geometry(filename) yields the data stored at the freesurfer geometry file given
      by filename. The optional argument 'to' may be used to change the kind of data that is
      returned.

    The following are valid settings for the 'to' keyword argument:
      * 'mesh' (the default) yields a mesh object
      * 'tess' yields a tess object (discarding coordinates)
      * 'raw' yields a tuple of numpy arrays, identical to the read_geometry return value.
    '''
    if not warn:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    category=UserWarning,
                                    module='nibabel')
            (xs, fs, info) = fsio.read_geometry(filename, read_metadata=True)
    else:
        (xs, fs, info) = fsio.read_geometry(filename, read_metadata=True)
    # see if there's chirality data here...
    filename = os.path.split(filename)[1]
    filename = filename.lower()
    if   filename.startswith('lh'): info['chirality'] = 'lh.'
    elif filename.startswith('rh'): info['chirality'] = 'rh.'
    # parse it into something
    to = to.lower()
    if to in ['mesh', 'auto', 'automatic']:
        return geo.Mesh(fs, xs, meta_data=info)
    elif to in ['tess', 'tesselation']:
        return geo.Tesselation(fs, meta_data=info)
    elif to in ['coords', 'coordinates']:
        return xs
    elif to in ['triangles', 'faces']:
        return fs
    elif to in ['meta', 'meta_data']:
        return info
    elif to =='raw':
        return (xs, fs)
    else:
        raise ValueError('Could not understand \'to\' argument: %s' % to)
@nyio.exporter('freesurfer_geometry', ('sphere.reg',))
def save_freesurfer_geometry(filename, obj, volume_info=None, create_stamp=None):
    '''
    save_mgh(filename, obj) saves the given object to the given filename in the mgh format and
      returns the filename.

    All options that can be given to the to_mgh function can also be passed to this function; they
    are used to modify the object prior to exporting it.
    '''
    obj = geo.to_mesh(obj)
    fsio.write_geometry(filename, obj.coordinates.T, obj.tess.faces.T,
                        volume_info=volume_info, create_stamp=create_stamp)
    return filename
@nyio.importer('freesurfer_morph', ('curv',))
def load_freesurfer_morph(filename):
    '''
    load_freesurfer_morph(filename) yields the result of loading the given filename as FreeSurfer
      morph-data (e.g., lh.curv).
    '''
    return fsio.read_morph_data(filename)
@nyio.exporter('freesurfer_morph', ('curv',))
def save_freesurfer_morph(filename, obj, face_count=0):
    '''
    save_freesurfer_morph(filename, obj) saves the given object using nibabel.freesurfer.io's
      write_morph_data function, and returns the given filename.
    '''
    fsio.write_morph_data(filename, obj, fnum=face_count)
    return filename
@nyio.importer('freesurfer_label', ('label',))
def load_freesurfer_label(filename, to='all'):
    '''
    load_freesurfer_label(filename) yields the boolean label property found in the given freesurfer
      label file.

    The return value is determined by the option `to`, which by default is 'all'. The following
    values may be given:
      * 'vertices' is equivalent to nibabel.freesurfer.io.read_label(filename) as is.
      * 'all' is equivalent to nibabel.freesurfer.io.read_label(filename, read_scalars=True) as is.
    '''
    if to in [None,Ellipsis]: to = 'all'
    to = to.lower()
    if to == 'vertices': fsio.read_label(filename, read_scalars=False)
    (ls,ps) = fsio.read_label(filename, read_scalars=True)
    if to == 'all': return (ls,ps)
    else: raise ValueError('could not parse to option: %s' % to)    
@nyio.importer('freesurfer_annot', ('annot',))
def load_freesurfer_annot(filename, to='property'):
    '''
    load_freesurfer_annot(filename) yields the result of loading an annotation file given by the
      provided filename.

    The precise return value is determined by the optional argument `to`; it is 'property' by
    default but may be set to any of the following:
      * 'raw': nibabel.freesurfer.io.read_annot(filename, orig_ids=False)
      * 'raw_orig': nibabel.freesurfer.io.read_annot(filename, orig_ids=True)
      * 'property': returns the label ids found in the file as a property array.
      * 'index': returns the label_index object stored in the annotation file.
      * 'all': returns (property, index).
    '''
    if to in [None,Ellipsis]: to = 'property'
    to = to.lower()
    if to == 'raw_orig': return fsio.read_annot(filename, orig_ids=True)
    dat = fsio.read_annot(filename, orig_ids=False)
    if to == 'raw': return dat
    (ls,clrs,nms) = dat
    nms = np.asarray([nm.decode('utf-8') for nm in nms])
    lbls = clrs[:,-1]
    clrs = clrs[:,:4]
    clrs[:,3] = 255 - clrs[:,3]
    clrs = clrs / 255.0
    oks = ls >= 0
    p = np.array(ls)
    p[~oks] = 0
    lils = np.arange(len(nms))
    if to in ['property', 'prop', 'p', 'auto', 'automatic']: return p
    elif to in ['index', 'label_index', 'lblidx', 'idx']: return label_index(lils, nms, clrs)
    elif to in ['all', 'full']: return (p, label_index(lils, nms, clrs))
    else: raise ValueError('bad to conversion: %s' % to)
@nyio.exporter('freesurfer_annot', ('annot',))
def save_freesurfer_annot(filename, obj, index=None):
    '''
    save_freesurfer_annot(filename, prop) saves the given integer property prop to the given
      filename as a FreeSurfer annotation file.

    The optional argument index specifies how the colortab and names of the property labels should
    be handles. By default this is None, in which case names are generated using the formatter
    'label%d' and color values are generated using the label_colors function. If index is not None,
    then it is coerced to an index using the to_label_index() function, and the label index is
    used to create the colortable and names.
    '''
    # first parse the index
    if index is None:
        if len(obj) == 2 and pimms.is_vector(obj[0]): (obj, index) = obj
        else: index = label_index(obj)
    index = to_label_index(index)
    # okay, let's get the data in the right format
    (u,ris) = np.unique(obj, return_inverse=True)
    es = [index[l] for l in u]
    clrs = np.round([np.asarray(e.color)*255 for e in es]).astype('int')
    clrs[:,3] = 255 - clrs[:,3] # alpha -> transparency
    nms  = [e.name for e in es]
    lbls = [e.id for e in es]
    fsio.write_annot(filename, ris, clrs, nms, fill_ctab=True)
    return filename

# A few annot labels we can just save:
brodmann_label_index = label_index(
    np.arange(15),
    ['none', 'BA1', 'BA2', 'BA3a', 'BA3b', 'BA4a', 'BA4p', 'BA6', 'BA44', 'BA45', 'V1', 'V2', 'MT',
     'perirhinal', 'enterorhinal'])
label_indices['freesurfer_brodmann'] = brodmann_label_index.persist()



####################################################################################################
# __init__.py

'''Tools for analyzing and registering cortical meshes.'''

submodules = ('neuropythy.util.conf',
              'neuropythy.util.core',
              'neuropythy.util.filemap',
              'neuropythy.util.labels',
              'neuropythy.util',
              'neuropythy.math.core',
              'neuropythy.math',
              'neuropythy.java',
              'neuropythy.io.core',
              'neuropythy.io',
              'neuropythy.geometry.util',
              'neuropythy.geometry.mesh',
              'neuropythy.geometry',
              'neuropythy.optimize.core',
              'neuropythy.optimize',
              'neuropythy.mri.core',
              'neuropythy.mri.images',
              'neuropythy.mri',
              'neuropythy.freesurfer.core',
              'neuropythy.freesurfer',
              'neuropythy.hcp.files',
              'neuropythy.hcp.core',
              'neuropythy.hcp',
              'neuropythy.registration.core',
              'neuropythy.registration',
              'neuropythy.vision.models',
              'neuropythy.vision.retinotopy',
              'neuropythy.vision.cmag',
              'neuropythy.vision',
              'neuropythy.graphics.core',
              'neuropythy.graphics',
              'neuropythy.datasets.core',
              'neuropythy.datasets.benson_winawer_2018',
              'neuropythy.datasets.hcp',
              'neuropythy.datasets.visual_performance_fields',
              'neuropythy.datasets.hcp_lines',
              'neuropythy.datasets',
              'neuropythy.plans.core',
              'neuropythy.plans.prfclean',
              'neuropythy.plans',
              'neuropythy.commands.surface_to_ribbon',
              'neuropythy.commands.benson14_retinotopy',
              'neuropythy.commands.register_retinotopy',
              'neuropythy.commands.atlas',
              'neuropythy.commands')
'''neuropythy.submodules is a tuple of all the sub-modules of neuropythy in a loadable order.'''

def reload_neuropythy():
    '''
    reload_neuropythy() reloads all of the modules of neuropythy and returns the reloaded
    neuropythy module. This is similar to reload(neuropythy) except that it reloads all the
    neuropythy submodules prior to reloading neuropythy.

    Example:
      import neuropythy as ny
      # ... some nonsense that breaks the library ...
      ny = ny.reload_neuropythy()
    '''
    import sys, six
    if not six.PY2:
        try:              from importlib import reload
        except Exception: from imp import reload
    for mdl in submodules:
        if mdl in sys.modules:
            sys.modules[mdl] = reload(sys.modules[mdl])
    return reload(sys.modules['neuropythy'])

from   .util       import (config, is_image, library_path, to_affine, is_address, address_data,
                           is_curve_spline, to_curve_spline, curve_spline, flattest,
                           is_list, is_tuple, to_hemi_str, is_dataframe, to_dataframe, auto_dict,
                           label_index, is_label_index, to_label_index)
from   .util       import label_indices as labels
from   .io         import (load, save, to_nifti)
from   .mri        import (is_subject, is_cortex, to_cortex, to_image, to_image_spec,
                           is_image_spec, image_interpolate, image_apply, image_copy, image_clear,
                           is_pimage, to_image_type)
from   .vision     import (retinotopy_data, empirical_retinotopy_data, predicted_retinotopy_data,
                           register_retinotopy, retinotopy_anchors, retinotopy_model,
                           neighborhood_cortical_magnification, as_retinotopy,
                           retinotopy_comparison, to_logeccen, from_logeccen)
from   .geometry   import (mesh, tess, topo, map_projection, path_trace, 
                           is_vset, is_mesh, is_tess, is_topo, is_flatmap, paths_to_labels,
                           is_map_projection, is_path, is_path_trace, close_path_traces,
                           to_mesh, to_tess, to_property, to_mask, to_flatmap, to_map_projection,
                           isolines, map_projections)
from   .freesurfer import (subject as freesurfer_subject, to_mgh)
from   .hcp        import (subject as hcp_subject)
from   .datasets   import data
from . import util
from . import math
from . import freesurfer
from . import hcp
from . import plans

# things we might want to load but that might fail
try:
    from .graphics import cortex_plot
    from .         import graphics
except Exception: pass

# Version information...
__version__ = '0.12.1'
